# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import base64
from collections import OrderedDict
import json
import logging

from iotcs.common import RestApi
from iotcs.common import Util


__query = {} \
    if Util.getConfig()["device_model"].getboolean("allow_draft_device_models") \
    else { 'createDraft' : 'false' }

#
# REST resource for activation policy
#
__REST_ACTIVATION_POLICY = RestApi.V2.getReqRoot() + "/activation/policy"
#
# REST resource for direct activation of a device
#
__REST_DIRECT_ACTIVATION = RestApi.V2.getReqRoot() + "/activation/direct"
#
# REST resource for indirect activation of a device
#
__REST_INDIRECT_ACTIVATION = RestApi.V2.getReqRoot() + "/activation/indirect/device"

class ActivationPolicy():

    PROPERTY_OS_NAME = "OSName"
    PROPERTY_OS_VERSION = "OSVersion"

    __slots__ = (
        'keyType',
        'keySize',
        'hashAlgorithm'
    )

    # TODO: To be consistent this should probably be done using fromJson method, not from "init"
    def __init__(self, activationPolicyResponse):
        """

        :param activationPolicyResponse:  is the raw response object from urllib
        """
        resp = str(activationPolicyResponse, 'utf-8')
        policyJson = json.loads(resp)
        self.keyType = policyJson['keyType']
        self.keySize = policyJson['keySize']
        self.hashAlgorithm = policyJson['hashAlgorithm']

    def getKeyType(self):
        return self.keyType

    def getKeySize(self):
        return self.keySize

    def getHashAlgorithm(self):
        return self.hashAlgorithm

DIRECT_ACTIVATION_URN =  "urn:oracle:iot:dcd:capability:direct_activation"
INDIRECT_ACTIVATION_URN = "urn:oracle:iot:dcd:capability:indirect_activation"
DEFAULT_MESSAGE_DIGEST_ALGORITHM = "HmacSHA256"

def getActivationPolicy(secureConnection, deviceId):
    """

    :param secureConnection:
    :param deviceId:
    :return: `ActivationPolicy`
    """

    query = dict()
    query[ActivationPolicy.PROPERTY_OS_NAME] = Util.getOSName().replace(' ', '+')
    query[ActivationPolicy.PROPERTY_OS_VERSION] = Util.getOSversion()

    # TODO: secureConnections should return HttpResponse
    policyResponse = secureConnection.get(__REST_ACTIVATION_POLICY, query)

    if policyResponse.getStatus() == 401:
        # TODO: Create Exceptions i.e. IllegalState, etc.
        raise ValueError(msg = "Endpoint already activated")

    if policyResponse.getStatus() != 200:
        raise ValueError(f'ActivationPolicy failed GET status: {policyResponse.getStatus()}')

    policy = policyResponse.getData()
    activationPolicy = ActivationPolicy(policy)

    # FINEST
    logging.getLogger('iolib.device.impl.ActivationManager').debug(f'getActivationPolicy: response - {str(policy)}')

    return activationPolicy

def createDirectActivationRequest(trustedAssetsManager, hashAlgorithm, deviceModels):
    """

    :param trustedAssetsManager:
    :param hashAlgorithm:
    :param deviceModels: Set of device model urns
    :return:
    """

    subjectPublicKeyInfo = DirectActivationRequest.SubjectPublicKeyInfo()
    certificationRequestInfo = DirectActivationRequest.CertificationRequestInfo()

    certificationRequestInfo.setSubjectPublicKeyInfo(subjectPublicKeyInfo)
    certificationRequestInfo.setSubject(trustedAssetsManager.getClientId())
    request = DirectActivationRequest()
    request.setCertificationRequestInfo(certificationRequestInfo)
    rdm = request.getDeviceModels()
    for m in deviceModels:
        rdm.add(m)
    __signRequest(request, trustedAssetsManager, hashAlgorithm)
    return request


def postDirectActivationRequest(secureConnection, directActivationRequest):
    """

    :param secureConnection:
    :param directActivationRequest:
    :return:
    """

    payload = directActivationRequest.toJson()

    # While the server supports this query parameter it is not required by the library
    # and the mock server does not support it
    #query = {'requestCertificate': 'true'}
    #for (k,v) in __query.items():
    #   query[k] = v

    # TODO: shouldn't this use the default "TIMEOUT" value ?
    resp = secureConnection.post(__REST_DIRECT_ACTIVATION, __query, payload)
    if resp.getStatus() != 200:
        raise ValueError(f'postDirectActivationRequest(): Failed to activate device, response.status: {resp.getStatus()}')

    # TODO: probably should 'pusd down' json conversion into fromJson methods
    resp_data = resp.getData()
    response_json = json.loads(resp_data, encoding='utf-8')
    directactivationresponse = DirectActivationResponse().fromJson(response_json)

    # FINEST
    logging.getLogger('iolib.device.impl.ActivationManager').debug(f'postDirectActivationRequest: response - {json.dumps(response_json)}')


    return directactivationresponse

def createIndirectActivationRequest(hardwareId, metadata, deviceModels, signature):
    request = IndirectActivationRequest(hardwareId, metadata, deviceModels, signature);
    return request

def postIndirectActivationRequest(secureConnection, indirectActivationRequest):
    """

    :param secureConnection:
    :param indirectActivationRequest:
    :return:
    """

    payload = indirectActivationRequest.toJson()
    resp = secureConnection.post(__REST_INDIRECT_ACTIVATION, __query, payload, timeout=-1)
    if resp.getStatus() == 401:
        raise ValueError("endpoint already activated")

    if resp.getStatus() != 200:
        logging.getLogger('iolib.device.impl.ActivationManager').debug(f'postIndirectActivationRequest: payload - {payload}')
        raise ValueError(f'Indirect activation failed: {resp.getVerboseStatus("POST", __REST_INDIRECT_ACTIVATION)}')

    # TODO: probably should 'pusd down' json conversion into fromJson methods
    resp_data = resp.getData()
    response_json = json.loads(resp_data, encoding='utf-8')
    indirectactivationresponse = IndirectActivationResponse().fromJson(response_json)

    # FINEST
    logging.getLogger('iolib.device.impl.ActivationManager').debug(f'postIndirectActivationRequest: response - {json.dumps(response_json)}')
    return indirectactivationresponse

def __signRequest(directActivationRequest, trustedAssetsManager, hashAlgorithm):

    certificationRequestInfo = directActivationRequest.getCertificationRequestInfo()
    subjectPublicKeyInfo = certificationRequestInfo.getSubjectPublicKeyInfo()

    publicKey = trustedAssetsManager.getPublicKey()
    # TODO: Can't seem to map these values from JCL to PCL there is not "PublicKey" object in PCL
    # TODO: There is a PublicKey object in cryptodome, not used in original code.
    # TODO: Could pass the activation policy keyType too to get 'RSA', if we can't get it from
    # TODO: the public key.
    subjectPublicKeyInfo.setAlgorithm('RSA') #publicKey.getAlgorithm())
    subjectPublicKeyInfo.setPublicKey(publicKey)
    # TODO: Need publicKey object to get encoding format.
    subjectPublicKeyInfo.setKeyFormat(DirectActivationRequest.SubjectPublicKeyInfo.PUBILIC_KEY_ENCODING_FORMAT_X509) #publicKey.getFormat())
    subjectPublicKeyInfo.setSecretHashAlgorithm(DEFAULT_MESSAGE_DIGEST_ALGORITHM)

    clientSecretData = bytes(trustedAssetsManager.getClientId(), 'utf-8')
    # TODO: Whey the distinction between "signWithSharedSecret" and "signeWithSharedSecretHardwareId" ?
    # TODO: "WithHardwareId" is "signWithSharedSecret" in JCL. Made "signWithSharedSecret" the same as JCL
    # TODO: Now works for icd's as well
    # TODO: Does not support passing the SecretHashAlgorithm. Pass None and the signWithSharedSecret will
    # TODO: default to SHA256 as in the original code
    clientSecret = trustedAssetsManager.signWithSharedSecret(clientSecretData, None, trustedAssetsManager.getClientId())

    # TODO: signWithPrivateKey hard codes the hashAlgorithm
    signaturePayload = __getSignaturePayload(certificationRequestInfo, clientSecret)
    signature = trustedAssetsManager.signWithPrivateKey(signaturePayload) #, hashAlgorithm

    signature = base64.b64encode(signature)

    directActivationRequest.setSignatureAlgorithm(hashAlgorithm)
    directActivationRequest.setSignature(signature)

def __getSignaturePayload(requestInfo, clientSecret):

    subjectPublicKeyInfo = requestInfo.getSubjectPublicKeyInfo()
    # TODO: Use better way to format this string.
    payload = f'''{requestInfo.getSubject()}
{subjectPublicKeyInfo.getAlgorithm()}
{subjectPublicKeyInfo.getKeyFormat()}
{subjectPublicKeyInfo.getSecretHashAlgorithm()}
'''
    # format is "\n<attribute>='<attributevalue>'" or '\n<attribute>="null"'
    map = requestInfo.getAttributes()
    if map != None:
        for (k,v) in map.items():
            if v == 'None':
                v = "null"
            else:
                v = f"'{v}'"
            payload = f'''{payload}\n{k}={v}'''


    payloadBytes = payload.encode('utf-8')
    signatureBytes = payloadBytes + clientSecret + subjectPublicKeyInfo.getPublicKey()

    return signatureBytes

class DirectActivationRequest():

    FIELD_DEVICE_MODELS = "deviceModels"
    FIELD_CERTIFICATION_REQUEST_INFO = "certificationRequestInfo"
    FIELD_SIGNATURE_ALGORITHM = "signatureAlgorithm"
    FIELD_SIGNATURE = "signature"

    class SubjectPublicKeyInfo():

        PUBILIC_KEY_ENCODING_FORMAT_X509 = 'X.509'
        FIELD_ALGORITHM = "algorithm"
        FIELD_PUBLIC_KEY = "publicKey"
        FIELD_FORMAT = "format"
        FIELD_SECRET_HASH_ALGORITHM = "secretHashAlgorithm"

        __slots__ = (
            'algorithm',
            'publickey',
            'secrethashalgorithm',
            'keyformat'
        )

        def __init__(self):
            self.keyformat = self.PUBILIC_KEY_ENCODING_FORMAT_X509
            self.algorithm = None
            self.publickey = None
            self.secrethashalgorithm = None

        def getAlgorithm(self):
            return self.algorithm

        def setAlgorithm(self, algorithm):
            self.algorithm = algorithm

        def getPublicKey(self):
            return self.publickey

        def setPublicKey(self, publickey):
            self.publickey = publickey

        def getKeyFormat(self):
            return self.keyformat

        def setKeyFormat(self, keyformat):
            self.keyformat = keyformat

        def getSecretHashAlgorithm(self):
            return self.secrethashalgorithm

        def setSecretHashAlgorithm(self, secrethashalgorithm):
            self.secrethashalgorithm = secrethashalgorithm

        def toJson(self):
            """

            :return: jsonObject (python object)
            """
            # TODO: why OrderedDict ?
            dct = OrderedDict()
            if self.algorithm is not None:
                dct[self.FIELD_ALGORITHM] = self.algorithm

            dct[self.FIELD_FORMAT] = self.keyformat
            if self.secrethashalgorithm is not None:
                dct[self.FIELD_SECRET_HASH_ALGORITHM] = self.secrethashalgorithm
            if self.publickey is not None:
                dct[self.FIELD_PUBLIC_KEY] = base64.b64encode(self.publickey).decode('utf-8')

            return dct

        def fromJson(self, jsonObject):
            if jsonObject is None:
                return None

            if self.FIELD_FORMAT in jsonObject:
                self.setKeyFormat(jsonObject[self.FIELD_FORMAT])
            if self.FIELD_ALGORITHM in jsonObject:
                self.setAlgorithm(jsonObject[self.FIELD_ALGORITHM])
            if self.FIELD_PUBLIC_KEY in jsonObject:
                encodedPublicKey = jsonObject[self.FIELD_PUBLIC_KEY]
                self.setPublicKey(base64.b64decode(encodedPublicKey))
            if self.FIELD_SECRET_HASH_ALGORITHM in jsonObject:
                self.setSecretHashAlgorithm(jsonObject[self.FIELD_SECRET_HASH_ALGORITHM])

        def __str__(self):
            return f'''SubjectPublicKeyInfo{{
                algorithm='{self.algorithm}'
                , publicKey={self.publickey}
                , keyformat='{self.keyformat}'
                , secretHashAlgorithm='{self.secrethashalgorithm}'}}'''

        # TODO: What is this ?
        def is_valid(self):
            raise NotImplementedError('SubjectPublicKeyInfo.is_valid is not implemented')


    class CertificationRequestInfo():

        FIELD_SUBJECT = "subject"
        FIELD_SUBJECT_PUBLIC_KEY_INFO = "subjectPublicKeyInfo"
        FIELD_ATTRIBUTES = "attributes"

        __slots__ = (
            'subject',
            'attributes',
            'subjectPublicKeyInfo'
        )

        def __init__(self):
            self.subject = None
            self.subjectPublicKeyInfo = None
            self.attributes = dict()

        def getSubject(self):
            return self.subject

        def setSubject(self, subject):
            self.subject = subject

        def getAttributes(self):
            return self.attributes

        def setAttributes(self, attributes):
            self.attributes = attributes

        def getSubjectPublicKeyInfo(self):
            return self.subjectPublicKeyInfo

        def setSubjectPublicKeyInfo(self, subjectPublicKeyInfo):
            self.subjectPublicKeyInfo = subjectPublicKeyInfo

        def toJson(self):
            # TODO: why OrderedDict ?
            dct = OrderedDict()
            dct[self.FIELD_SUBJECT] = self.subject
            if self.subjectPublicKeyInfo is not None:
                dct[self.FIELD_SUBJECT_PUBLIC_KEY_INFO] = self.subjectPublicKeyInfo.toJson()

            dct[self.FIELD_ATTRIBUTES] = self.attributes

            # TODO: Can any of these statements raise exception ?
            return dct

        def fromJson(self, jsonObject):
            if jsonObject is None:
                return None

            if self.FIELD_SUBJECT in jsonObject:
                self.setSubject(jsonObject[self.FIELD_SUBJECT])
            if self.FIELD_SUBJECT_PUBLIC_KEY_INFO in jsonObject:
                subjectPublicKeyInfo = DirectActivationRequest.SubjectPublicKeyInfo()
                subjectPublicKeyInfoJsonObj = jsonObject[self.FIELD_SUBJECT_PUBLIC_KEY_INFO]
                if subjectPublicKeyInfoJsonObj is not None:
                    subjectPublicKeyInfo.fromJson(subjectPublicKeyInfoJsonObj)

            if self.FIELD_ATTRIBUTES in jsonObject:
                self.setAttributes(jsonObject[self.FIELD_ATTRIBUTES])


        def __str__(self):
            return f'''CertificationRequestInfo{{
                subject='{self.subject}'
                , subjectPublicKeyInfo={self.subjectPublicKeyInfo}
                , attributes={self.attributes}}}'''


    __slots__ = (
        'certificationRequestInfo',
        'signatureAlgorithm',
        'signature',
        'deviceModels'
    )

    def __init__(self):
        self.deviceModels = set()
        self.certificationRequestInfo = None
        self.signatureAlgorithm = None
        self.signature = None

    def getCertificationRequestInfo(self):
        return self.certificationRequestInfo

    def setCertificationRequestInfo(self, certificationRequestInfo):
        self.certificationRequestInfo = certificationRequestInfo

    def getSignatureAlgorithm(self):
        return self.signatureAlgorithm

    def setSignatureAlgorithm(self, signatureAlgorithm):
        self.signatureAlgorithm = signatureAlgorithm

    def setSignature(self, signature):
        self.signature = signature

    def getDeviceModels(self):
        return self.deviceModels

    @staticmethod
    def fromJson(jsonString):
        try:
            dar = DirectActivationRequest()
            jsonObject = json.loads(jsonString)
            return dar._fromJson(jsonObject)
        except Exception as e:
            raise ValueError(f'Invalid json {jsonString}: {e}')

    # TODO: Distinguish between the 2 "fromJson" and why need both ?
    def _fromJson(self, jsonObject):
        fieldJsonObject = None
        if DirectActivationRequest.FIELD_CERTIFICATION_REQUEST_INFO in jsonObject:
            self.certificationRequestInfo = DirectActivationRequest.CertificationRequestInfo()
            fieldJsonObject = jsonObject[DirectActivationRequest.FIELD_CERTIFICATION_REQUEST_INFO]
            self.certificationRequestInfo.fromJson(fieldJsonObject)

        if DirectActivationRequest.FIELD_SIGNATURE_ALGORITHM in jsonObject:
            encodedSignature = jsonObject[DirectActivationRequest.FIELD_SIGNATURE_ALGORITHM]
            self.setSignature(base64.b64decode(jsonObject[DirectActivationRequest.FIELD_SIGNATURE_ALGORITHM], validate=True))

        if DirectActivationRequest.FIELD_DEVICE_MODELS in jsonObject:
            deviceModels = jsonObject[DirectActivationRequest.FIELD_DEVICE_MODELS]
            for m in deviceModels:
                self.deviceModels.add(m)

    def toJson(self):
        # TODO: Why OrderedDict ?
        jsonObject = OrderedDict()
        try:
            if self.certificationRequestInfo is not None:
                jsonObject[DirectActivationRequest.FIELD_CERTIFICATION_REQUEST_INFO] = self.certificationRequestInfo.toJson()

            if self.signatureAlgorithm is not None:
                jsonObject[DirectActivationRequest.FIELD_SIGNATURE_ALGORITHM] = self.signatureAlgorithm

            if self.signature is not None:
                jsonObject[DirectActivationRequest.FIELD_SIGNATURE] = str(self.signature, 'utf-8')

            if self.deviceModels is not None and len(self.deviceModels) > 0:
                jsonObject[DirectActivationRequest.FIELD_DEVICE_MODELS] = list(self.deviceModels)
        except Exception as e:
            raise ValueError(f'Cannot convert DirectActivationRequest to json object: {e}')

        return jsonObject

    def __str__(self):
        return f'''DirectActivationRequest{{
                deviceModels={str(self.deviceModels) if len(self.deviceModels) > 0 else ""}
                , certificationRequestInfo={self.certificationRequestInfo}
                , signatureAlgorithm='{self.signatureAlgorithm}'
                , signature={self.signature}}}'''

class DirectActivationResponse():
    """
    DirectActivationResponse
    """

    FIELD_ENDPOINT_ID = "endpointId"
    FIELD_ACTIVATION_TIME = "activationTime"
    FIELD_ENDPOINT_STATE = "endpointState"
    FIELD_CERTIFICATE = "certificate"

    __slots__ = (
        'endpointId',
        'activationTime',
        'endpointState',
        'certificate'
    )

    def getCertificate(self):
        return self.certificate

    def setCertificate(self, certificate):
        self.certificate = certificate

    def __init__(self):
        self.endpointId = None
        self.activationTime = None
        self.endpointState = None
        self.certificate = []

    def getActivationTime(self):
        return self.activationTime

    def setActivationTime(self, activationTime):
        self.activationTime = activationTime

    def getEndpointState(self):
        return self.endpointState

    def setEndpointState(self, endpointState):
        self.endpointState = endpointState

    def getEndpointId(self):
        return self.endpointId

    def setEndpointId(self, endpointId):
        self.endpointId = endpointId

    def __str__(self):
        return f'''ActivationResponse{{endpointId='{self.endpointId}'
            , activationTime={self.activationTime}
            , endpointState='{self.endpointState}'
            , certificate={self.certificate}}}'''

    def fromJson(self, jsonObject):
        response = DirectActivationResponse()
        try:
            response.setEndpointId(jsonObject[self.FIELD_ENDPOINT_ID])
            #  TODO: implement a Date converter
            # activationTime =
            #  getJsonDate(jsonObject.get(FIELD_ACTIVATION_TIME));
            response.setEndpointState(jsonObject[self.FIELD_ENDPOINT_STATE])
        except Exception as e:
            raise ValueError(f'DirectActivationResponse.fromJson(): {e}')

        #  TODO: we're manually replacing EOLs here, need to figure out a
        #  better solution
        fieldCertificate = None
        certificateString = None
        if self.FIELD_CERTIFICATE in jsonObject:
            fieldCertificate = jsonObject[self.FIELD_CERTIFICATE]
            if fieldCertificate != None:
                certificateString = fieldCertificate.__str__().replace("\r", "").replace("\n", "")
            # TODO: Does this produce an Array ?
            response.certificate = base64.b64decode(certificateString)
        else:
            response.certificate = []
        return response

    def toJson(self):
        jsonObject = dict()
        try:
            jsonObject[self.FIELD_ENDPOINT_ID] = self.endpointId
            jsonObject[self.FIELD_ENDPOINT_STATE] = self.endpointState
            jsonObject[self.FIELD_ACTIVATION_TIME] = self.activationTime.__str__()
        except Exception as e:
            raise ValueError(f'DirectActivationResponse,toJson(): failed {e}')

        return jsonObject

class IndirectActivationResponse():

    __slots__ = (
        'endpointId',
        'activationTime',
        'endpointState'
    )

    FIELD_ENDPOINT_ID = "endpointId"
    FIELD_ACTIVATION_TIME = "activationTime"
    FIELD_ENDPOINT_STATE = "endpointState"

    # TODO: make all classes consistent, accept all fields with None as default
    # TODO: fromJson, toJson, instance methods with use case IndirectActivationResponse().fromJson() etc.
    def __init__(self):
        self.endpointId = None
        self.activationTime = None
        self.endpointState = None

    def getEndpointId(self):
        return self.endpointId

    def setEndpointId(self, endpointId):
        self.endpointId = endpointId

    def getActivationTime(self):
        return self.activationTime

    def setActivationTime(self, activationTime):
        self.activationTime = activationTime

    def getEndpointState(self):
        return self.endpointState

    def setEndpointState(self, endpointState):
        self.endpointState = endpointState

    def toJson(self):
        jsonObject = OrderedDict()
        endpointId = self.getEndpointId()
        if endpointId is not None:
            jsonObject[self.FIELD_ENDPOINT_ID] = endpointId
        endpointState = self.getEndpointState()
        if endpointState is not None:
            jsonObject[self.FIELD_ENDPOINT_STATE] = endpointState
        activationTime = self.activationTime()
        if activationTime is not None:
            jsonObject[self.FIELD_ACTIVATION_TIME] = activationTime

        return jsonObject

    def fromJson(self, jsonObject):
        """

        :param jsonObject:  python object
        :return: this instance
        """
        if jsonObject is None:
            return

        for key in jsonObject.keys():
            if key == self.FIELD_ENDPOINT_STATE:
                self.setEndpointState(jsonObject[key])
            elif key == self.FIELD_ACTIVATION_TIME:
                self.setActivationTime(jsonObject[key])
            elif key == self.FIELD_ENDPOINT_ID:
                self.setEndpointId(jsonObject[key])

        return self

    def __str__(self):
        return json.dumps(self.toJson())

class IndirectActivationRequest:

    __slots__ = (
        'hardwareId', # the device hardware id for this activation request
        'metadata', # the device metadata
        'deviceModels', # the device model urn's for this device
        'signature', # the device signature
        'publicKey' # the request public key
    )

    FIELD_DEVICE_MODELS = "deviceModels"
    FIELD_SIGNATURE = "signature"
    HARDWARE_ID = "hardwareId";

    def __init__(self, hardwareId=None, metadata=None, deviceModels=None, signature=None):
        self.hardwareId = hardwareId
        if metadata is None:
            metadata = dict()
        self.metadata = metadata
        self.deviceModels = deviceModels
        self.signature = signature

    def getHardwareId(self):
        return self.hardwareId

    def setHardwareId(self, hardwareId):
        self.hardwareId = hardwareId

    def getDeviceModesls(self):
        return self.deviceModels

    def setDeviceModesl(self, deviceModels):
        self.deviceModels = deviceModels

    def getMetadata(self):
        return self.metadata

    def getSignature(self):
        return self.signature

    def getPublicKey(self):
        return self.publicKey

    def toJson(self):
        """
        Return a json object (python object)
        :return: json object (python object)
        """
        if self.hardwareId is None:
            raise ValueError("Hardware ID is None")

        dct = OrderedDict()
        if self.deviceModels:
            dct[self.FIELD_DEVICE_MODELS] = self.deviceModels

        if self.signature is not None:
            encodedsig = base64.b64encode(self.signature)
            dct[self.FIELD_SIGNATURE] = str(encodedsig, 'utf-8')

        for key, val in self.metadata.items():
            dct[key] = val

        dct[self.HARDWARE_ID] = self.hardwareId

        #jsonObject = json.loads(json.dumps(dct))

        return dct

    # Typical use is IndirectActivationRequest().fromJson(jsonObject)
    def fromJson(self, jsonObject):
        if jsonObject is None:
            # Should this throw ?
            return

        for key in jsonObject.keys():
            if key == self.HARDWARE_ID:
                self.setHardwareId(jsonObject[self.HARDWARE_ID])
            else:
                self.metadata[key] = jsonObject[key]

            if key == self.FIELD_DEVICE_MODELS:
                dmSet = set()
                # array of device models
                for dm in jsonObject[key]:
                    dmSet.add(dm)
                self.setDeviceModesl(dmSet)

            if key == self.FIELD_SIGNATURE:
                # TODO: confirm that we are convering bytes to string
                self.signature = base64.b64decode(jsonObject[key], 'utf-8')


    def __str__(self):
        return json.dumps(self.toJson())

