# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import collections
import ipaddress
import json
import logging
import netifaces
import shutil
import threading
import time

from iotcs.common import TimeManager
from iotcs.common import StatusCode
from iotcs.common import Util
from iotcs.message import DataMessage
from iotcs.message import Resource
from iotcs.message import ResponseMessage
from iotcs.messaging.client.device.util import RequestHandler

# com.oracle.iot.client.impl

class Diagnostics(RequestHandler):
    """
    The diagnostics :class:`.RequestHandler`.

    The :class:`iotcs.client.impl.Diagnostics` processes diagnostic
    :class:`iotcs.shared.message.RequestMessage`s.
    Diagnostics information includes total and free disk space, IP address,
    MAC address, diagnostics start time and client version.
    See `iotcs.config.config.py` section *diagnostics* and properties
    *client_paths*, *client_version*, *client_ip_address*, and
    *client_mac_address*.
    """

    __slots__ = (
        '__startTime',
        '__version',
        '__fixedIpAddress',
        '__fixedMacAddress'
    )
    DIAGNOSTICS_URN = "urn:oracle:iot:dcd:capability:diagnostics"
    DIAGNOSTICS_URL = \
        "urn:oracle:iot:dcd:capability:diagnostics:diagnostics/info"

    CONFIG_DIAGNOSTICS_SECTION = "diagnostics"
    CONFIG_CLIENT_VERSION_PROPERTY = "client_version"
    CONFIG_CLIENT_IP_ADDRESS_PROPERTY = "client_ip_address"
    CONFIG_CLIENT_MAC_ADDRESS_PROPERTY = "client_mac_address"
    CONFIG_CLIENT_PATHS_PROPERTY = "client_paths"

    def __init__(self):
        """Create a :class:`iotcs.client.impl.Diagnostics` instance."""
        super(Diagnostics, self).__init__()
        self.__startTime = TimeManager.currentTimeMillis()

        config = Util.getConfig()
        self.__version = config.get(Diagnostics.CONFIG_DIAGNOSTICS_SECTION,
                                    Diagnostics.CONFIG_CLIENT_VERSION_PROPERTY,
                                    fallback=None)
        self.__fixedIpAddress = \
            config.get(Diagnostics.CONFIG_DIAGNOSTICS_SECTION,
                       Diagnostics.CONFIG_CLIENT_IP_ADDRESS_PROPERTY,
                       fallback=None)
        self.__fixedMacAddress = \
            config.get(Diagnostics.CONFIG_DIAGNOSTICS_SECTION,
                       Diagnostics.CONFIG_CLIENT_MAC_ADDRESS_PROPERTY,
                       fallback=None)

    def handleRequest(self, request):
        """
        Handle a diagnostics :class:`iotcs.shared.message.RequestMessage`.

        The total and free disk space, IP address, MAC address, diagnostics
        start time and client version are returned.

        See :class:`iotcs.client.device.util.RequestHandler`

        :return: :class:`iotcs.shared.message.ResponseMessage` with the
            free disksapce, total diskspace, IP address, MAC address, the
            diagnostics start time, and client version.
        """
        if request is None:
            raise ValueError(f'Diagnostics.handlRequest(): request is None')

        statusCode = StatusCode.BAD_REQUEST
        response = ResponseMessage(requestMessage=request)
        method = request.getMethod().upper() \
            if request.getMethod() is not None else None
        if method == Resource.Method.GET.name.upper():
            try:
                job = dict()
                # JCL implements this method only for one path '.', this
                # library supports more if necessary but it's not clear if the
                # receiver of the response knows that.
                # usage is a 'Freespace' named tuplet [ 'path', 'free' ]
                usage = self.getFreeDiskSpace()
                freespace = None
                # special case one entry
                if len(usage) == 1:
                    freespace = usage[0].free
                elif len(usage) > 1:
                    freespace = []
                    for free in usage:
                        space = {free.path, free.free}
                        freespace.append(space)

                # TODO: What is the spec, if freespace cannot be determined
                if freespace is not None:
                    job["freeDiskSpace"] = freespace

                #  diagnostics rest API definition currently only
                #  supports a single address, should consider extending
                #  this to include an array of addresses
                #  (for multiple interfaces)
                if self.__fixedIpAddress is None:
                    # returns a tuple ifaceAndIp[0] is the iface, and
                    # ifaceAndIp[1] is the ip
                    ifaceAndIp = self.getIpAddress()
                    job["ipAddress"] = ifaceAndIp[1]
                    job["macAddress"] = self.getMacAddress(ifaceAndIp[0],
                                                           ifaceAndIp[1])
                else:
                    job["ipAddress"] = self.__fixedIpAddress
                    job["macAddress"] = self.__fixedMacAddress
                job["startTime"] = self.getStartTimeMs()

                # JCL implements this method only for one path '.', this
                # library supports more if necessary but it's not clear if the
                # receiver of the response knows that.
                # usage is a 'Totalspace' named tuplet [ 'path', 'total' ]
                usage = self.getTotalDiskSpace()
                totalspace = None
                # special case one entry
                if len(usage) == 1:
                    totalspace = usage[0].total
                elif len(usage) > 1:
                    totalspace = []
                    # usage is an array of Totalspace named tuple
                    for total in usage:
                        space = {total.path, total.total}
                        totalspace.append(space)
                if totalspace is not None:
                    job["totalDiskSpace"] = totalspace

                job["version"] = self.__version
                response.setBody(job)
                statusCode = StatusCode.OK
            except Exception as e:
                statusCode = StatusCode.BAD_REQUEST

        else:
            statusCode = StatusCode.METHOD_NOT_ALLOWED

        response.setStatusCode(statusCode)
        return response

    def getIpAddress(self):
        """
        Return a network interface and IP address.

        Return a tuple containing *(<network interface>, <ipaddress>)*.
        This first global IPv4 address is returned in the *AF_INET* family.
        If there are no IPv4 addresses, then the first IPv6 address.
        In no IP address can be determined, the tuple
        *( 'Unknown', 'Unknown' )* is returned.
        :return: ( <network interface>, <ipaddress> ) or
        ( 'Unknown', 'Unknown' )
        :raises:
        """
        here = self.getIpAddress.__qualname__

        # TODO: See collections named tuples

        ipAddress = None
        netIface = None

        nis = None
        try:
            # ['lo0', 'gif0', 'stf0', 'en0', 'en1', 'fw0']
            nis = netifaces.interfaces()
        except Exception as ignored:
            self.getLogger().warning(f'{here}: netifaces.interfaces() '
                                     f'raised exception {ignored}')
            return ("Unknown", "Unknown")
        if nis is None or len(nis) == 0:
            self.getLogger().debug(f'{here}: netifaces.interfaces() '
                                   f'returned None or empty list')
            return ("Unknown", "Unknown")

        # itf in ['lo0', 'gif0', 'stf0', 'en0', 'en1', 'fw0']
        for itf in nis:
            try:
                addresses = netifaces.ifaddresses(itf)
            except Exception as ignored:
                self.getLogger().warning(
                    f'{here}: (for ipv4) {itf}.ifaddresses() '
                    f'raised exception {ignored}')
                continue

            if addresses is None or len(addresses) == 0:
                self.getLogger().debug(
                    f'{here}: (for ipv4) {itf}.ifaddresses() '
                    f'returned None or empty list')
                continue

            # addresses is a dict of address families
            # look through ipv4 addresses
            if netifaces.AF_INET in addresses:
                # TODO: ipaddress has "is_global" should we use that instead ?
                # a list of dict's, each dict represents an address
                for ipv4addresses in addresses[netifaces.AF_INET]:
                    # if address.isLoopback() or address.isVirtual():
                    # loopback address
                    address = ipv4addresses["addr"]
                    ipv4address = ipaddress.IPv4Address(address)
                    if not ipv4address.is_loopback:
                        if not ipv4address.is_link_local:
                            # !loopBack && !linkLocal = global
                            ipAddress = address
                            netIface = itf
                            # we know it's ipv4, so we're done
                            # prefer first global
                            self.getLogger().debug(
                                f'{here}: have global ipv4 '
                                f'iface {netIface} and ip {ipAddress}')
                            break
                if ipAddress is not None:
                    break

        if ipAddress is not None:
            return (netIface, ipAddress)

        # Now try ipv6
        for itf in nis:
            try:
                addresses = netifaces.ifaddresses(itf)
            except Exception as ignored:
                self.getLogger().warning(
                    f'{here}: (for ipv6) {itf}.ifaddresses() '
                    f'raised exception {ignored}')
                continue

            if addresses is None or len(addresses) == 0:
                self.getLogger().debug(
                    f'{here}: (for ipv6) {itf}.ifaddresses() '
                    f'returned None or empty list')
                continue
            if netifaces.AF_INET6 in addresses:
                # a list of dict's, each dict represents an address
                for ipv6addresses in addresses[netifaces.AF_INET6]:
                    # if address.isLoopback() or address.isVirtual():
                    # loopback address
                    address = ipv6addresses["addr"]
                    ipv6address = ipaddress.IPv6Address(address)
                    if not ipv6address.is_loopback:
                        if not ipv6address.is_link_local:
                            # !loopBack && !linkLocal = global
                            # prefer first global
                            ipAddress = address
                            netIface = itf
                            # settle for ipv6
                            self.getLogger().debug(
                                f'{here}: have global ipv6 '
                                f'iface {netIface} and ip {ipAddress}')
                            break
                if ipAddress is not None:
                    break

        return (netIface, ipAddress)

    def getMacAddress(self, networkInterface, ipAddress):
        """
        Return the MAC address of `ipAddress`'s `networkInterface`.

        The MAC address of the first global `ipAddress`s `networkInterface`
        is returned. On a device that has multiple network interfaces, it is
        not defined which MAC address is returned.
        :param networkInterface: a netork interface
        :return: the MAC address for `ipAddress` on `networkInterface`
        """
        if networkInterface is None or networkInterface is "Unknown":
            return "Unknown"

        here = self.getMacAddress.__qualname__
        try:
            addresses = netifaces.ifaddresses(networkInterface)
        except Exception as ignored:
            self.getLogger().warning(
                f'{here}: {networkInterface}.ifaddresses()'
                f'raised exception {ignored}')
            return "Unknown"

        if addresses is None or len(addresses) == 0:
            self.getLogger().debug(f'{here}: {networkInterface}.ifaddresses()'
                                   f'returned None or empty list')
            return "Unknown"

        # See if the ip is ipv6 or ipv4
        inetfamily = netifaces.AF_INET
        if isinstance(ipaddress.ip_interface(ipAddress),
                      (ipaddress.IPv6Address)):
            self.getLogger().debug(f'{here}: have ipv6 address {ipAddress}')
            inetfamily = netifaces.AF_INET6

        if netifaces.AF_LINK not in addresses or \
                len(addresses[netifaces.AF_LINK]) == 0:
            self.getLogger().warning(
                f'{here}: net interface {networkInterface} '
                f'with ip {ipAddress} has no AF_LINK addresses')
            return "Unknown"

        if inetfamily not in addresses or len(addresses[inetfamily]) == 0:
            self.getLogger().warning(
                f'{here}: net interface {networkInterface} '
                f'with ip {ipAddress} has no family {inetfamily}')
            return "Unknown"

        # TODO: The example I got this from used the "addresses[0]".
        # TODO: Will there ever only be one ? experimentation says yes
        # TODO: (well on this machine)
        # TODO: Which is probably true, how many sets of addresses can you
        # TODO: have for a given 'en0' ?
        address = addresses[inetfamily][0]
        if_ip = address["addr"] if "addr" in address and \
                                   ipAddress == address["addr"] else None
        if if_ip is None:
            self.getLogger().warning(
                f'{here}: net interface {networkInterface} '
                f'with ip {ipAddress} has no match')
            return "Unknown"

        address = addresses[netifaces.AF_LINK][0]
        if_mac = address["addr"] if "addr" in address else None
        if if_mac is None:
            self.getLogger().warning(
                f'{here}: net interface {networkInterface} with '
                f'ip {ipAddress} has no "addr" address')
            return "Unknown"

        return if_mac

    def getFreeDiskSpace(self):
        """
        Return the amount of free disk space.

        See `iotcs.config.config.py` section *diagnostics* property
        *client_paths*. If this property is not set, the total disk space of
        *.* is returned.
        :return: list of namedtuple `Freespace`, [ 'path', 'free' ] or empty
        list if free space cannot be determined
        """
        # paths is ':' separated string of paths
        paths = \
            Util.getConfig().get(Diagnostics.CONFIG_DIAGNOSTICS_SECTION,
                                 Diagnostics.CONFIG_CLIENT_PATHS_PROPERTY,
                                 fallback='.')
        paths = paths.split(':')
        freespace = []
        Freespace = collections.namedtuple('Freespace', ['path', 'free'])
        # diskusage is a dict shutil.disk_usage named tuples [ 'free', 'total',
        # 'used' ]
        # where the key is a path from CONFIG_CLIENT_PATHS_PROPERTY
        diskusage = self.getDiskUsage(paths)
        for k, v in diskusage.items():
            freespace.append(Freespace(path=k, free=v.free))

        return freespace

    def getTotalDiskSpace(self):
        """
        Return the total disk space.

        See `iotcs.config.config.py` section *diagnostics* property
        *client_paths*. If this property is not set, the total disk space of
        *.* is returned.
        :return: list of namedtuple `Totalspace`, [ 'path', 'total' ] or empty
        list if total space cannot be determined
        """
        # paths is ':' separated string of paths
        paths = \
            Util.getConfig().get(Diagnostics.CONFIG_DIAGNOSTICS_SECTION,
                                 Diagnostics.CONFIG_CLIENT_PATHS_PROPERTY,
                                 fallback='.')
        paths = paths.split(':')
        totalspace = []
        Totalspace = collections.namedtuple('Totalspace', ['path', 'total'])
        # diskusage is a dict shutil.disk_usage named tuples [ 'free', 'total',
        # 'used' ] where the key is a path from CONFIG_CLIENT_PATHS_PROPERTY
        diskusage = self.getDiskUsage(paths)
        for k, v in diskusage.items():
            totalspace.append(Totalspace(path=k, total=v.total))

        return totalspace

    def getDiskUsage(self, paths):
        """
        Return total disk usage.

        :param paths: collect disk information from this list of paths
        :return: a list of namedtuple 'Diskusage'
            [ 'path', 'free', 'total', 'used' ]
        """
        here = self.getDiskUsage.__qualname__
        usage = dict()
        for path in paths:
            try:
                pathusage = shutil.disk_usage(path)
                usage[path] = pathusage
            except Exception as e:
                self.getLogger().warning(f'{here}: caught exception {e}')
        return usage

    def getStartTimeMs(self):
        """
        Return the diagnostics start time in milliseconds.

        :return: Return the diagnostics start time in milliseconds.
        """
        return self.__startTime

    def getLogger(self):
        """
        Return logger for :class:`iotcs.client.impl.Diagnostics`.

        :return: logger for :class:`iotcs.client.impl.Diagnostics`
        """
        return logging.getLogger("iotcs.client.impl.Diagnostics")


class DiagnosticCounters:
    def __init__(self):
        self.load = None
        self.totalBytesSent = None
        self.totalBytesReceived = None
        self.totalMessagesReceived = None
        self.totalMessagesRetried = None
        self.totalMessagesSent = None
        self.totalProtocolErrors = None


class TestConnectivity():
    """
    Handle diagnostics test connectivity messages.

    It is assumed there will only be one instance of this class constructed.
    """

    TEST_CONNECTIVITY_URL = "deviceModels/" + Diagnostics.DIAGNOSTICS_URN + \
                            "/testConnectivity"

    __slots__ = (
        # The number of Messages to send.
        'count',
        # The number of Messages already sent.
        'currentCount',
        # The interval between sending each Message.
        'interval',
        # The size of the payload of the Message.
        'size',
        # The ID of the current or last test connectivity invocation so we
        # can get the status and cancel.
        'messageDispatcher',
        'endpointId',
        # The status of the current or last test connectivity invocation.
        # We only keep one
        'testConnectivityThread',
        'testConnectivityHandler',
        'testConnectivityThreadLock'
    )

    def __init__(self, endpointId, messageDispatcher):
        """
        Create a :class:`.TestConnectivity` instance.

        :param endpointId: the endpoint ID of the device.
        :param messageDispatcher: a reference to a
            :class:`iotcs.client.device.MessageDispatcher`.
        """
        self.endpointId = endpointId
        self.messageDispatcher = messageDispatcher
        self.count = 0
        self.currentCount = 0
        self.interval = 0
        self.size = 0
        self.testConnectivityThread = None
        self.testConnectivityHandler = None
        self.testConnectivityThreadLock = threading.Lock()

    @classmethod
    def getAlreadyRunningResponse(cls, requestMessage):
        """
        Return :class:`.StatusCode.CONFLICT` response.

        Return a :class:`iotcs.shared.message.ResponseMessage` with
        :class:`.StatusCode.CONFLICT`
        :param requestMessage: the request message
        :return: a :class:`iotcs.shared.message.ResponseMessage` with
        :class:`.StatusCode.CONFLICT`
        """
        return cls.getResponseMessage(requestMessage,
                                      "Test connectivity is already running.",
                                      StatusCode.CONFLICT)

    @classmethod
    def getBadRequestResponse(cls, requestMessage, message):
        """
        Return a :class:`.StatusCode.BAD_REQUEST` response.

        Return a :class:`iotcs.shared.message.ResponseMessage` with
        :class:`.StatusCode.BAD_REQUEST`

        :param requestMessage: the request message
        :param message: the reason for the `BAD_REQUEST`
        :return: a :class:`iotcs.shared.message.ResponseMessage` with
            :class:`.StatusCode.BAD_REQUEST`
        """
        return cls.getResponseMessage(requestMessage, message,
                                      StatusCode.BAD_REQUEST)

    @classmethod
    def getResponseMessage(cls, requestMessage, body, statusCode):
        """
        Return a :class:`.ResponseMessage` with the given parameters.

        :param requestMessage: the request message
        :param body: the body of the response.
        :param  statusCode: the status code of the response.
        :return: a :class:`iotcs.shared.message.ResponseMessage` with the
            given parameters
        """
        return ResponseMessage(requestMessage=requestMessage,
                               body=body, statusCode=statusCode)

    @classmethod
    def getMethodNotAllowedResponse(cls, requestMessage):
        """
        Return :class:`.StatusCode.METHOD_NOT_ALLOWED` response.

        Return a :class:`iotcs.shared.message.ResponseMessage` with
        :class:`.StatusCode.METHOD_NOT_ALLOWED`

        :param requestMessage: the request message
        :return: a :class:`iotcs.shared.message.ResponseMessage` with
            :class:`.StatusCode.METHOD_NOT_ALLOWED`
        """
        return cls.getResponseMessage(
            requestMessage,
            bytearray(f'Unsupported request: {requestMessage.__str__()}',
                      encoding='utf-8'),
            StatusCode.METHOD_NOT_ALLOWED)

    @classmethod
    def getNullRequestResponse(cls, requestMessage):
        """
        Return :class:`.StatusCode.METHOD_NOT_ALLOWED` response.

        Return a :class:`iotcs.shared.message.ResponseMessage` with
        :class:`.StatusCode.METHOD_NOT_ALLOWED`
        :param requestMessage: the request message
        :return: a :class:`iotcs.shared.message.ResponseMessage` with
        :class:`.StatusCode.METHOD_NOT_ALLOWED`
        """
        return ResponseMessage(requestMessage=requestMessage,
                               body=b'Unsupported request.',
                               statusCode=StatusCode.METHOD_NOT_ALLOWED)

    @classmethod
    def getParam(cls, jsonObject, paramName, errors):
        """
        Return the value of `paramName` and accumulated errors.

        Return a tuple *(parmavalue, errors)*. The errors member is a
        concatenated string of accumulated errors.
        :param jsonObject: the json object containing `paramName`
        :param paramName: the json object attribute key
        :param errors: the error to include in the response if `paramName` is
        invvalid
        :return: a tuple *(paramvalue, errors)*
        """
        value = None
        try:
            if paramName in jsonObject and jsonObject[paramName] is not None:
                jsonNumber = jsonObject[paramName]
                try:
                    value = int(jsonNumber)
                    if value < 1:
                        errors = \
                            f'{errors} must be a numeric value greater then 0.'
                except Exception as nfe:
                    errors = \
                        f'{errors} must be a numeric value greater then 0.'
            else:
                errors = f'{errors} The {paramName} value must be supplied.'
        except Exception as cce:
            errors = f'{errors} "The {paramName} value must be a number.'
        return (value, errors)

    def getTestConnectivityHandler(self):
        """Return an instance of :class:`.TestConnectivityHandler`."""
        if self.testConnectivityHandler is None:
            self.testConnectivityHandler = \
                TestConnectivity.TestConnectivityHandler(self)
        return self.testConnectivityHandler

    def resetParameters(self):
        """Reset all parameters to 0."""
        self.count = 0
        self.currentCount = 0
        self.interval = 0
        self.size = 0

    class TestConnectivityHandler(RequestHandler):
        """
        The test connectivity :class:`.RequestHandler` class.

        The :class:`iotcs.client.impl.TestConnectivityHandler` processes
        test connectivity :class:`iotcs.shared.message.RequestMessage`s.
        The :class:`iotcs.client.impl.TestConnectivityHandler` handles
        *GET* and *PUT* :class:`iotcs.shared.message.RequestMessage`s.
        *GET* returns information about the state of the
        :class:`.TestConnectivityThread` including
        * *active* - whether the test connectivity thread is running
        * *count* - the number of pings to send to the IoT server
        * *currentCount* - the number pings sent
        * *interval* - the ping frequency in milliseconds
        * *size* - the size in bytes of the message to ping the server with

        *PUT* start a test connectivity thread if one is not in progress
        providing the *count*, *interval* and *size* parameters.
        """

        __slots__ = (
            'testConnectivity'
        )

        def __init__(self, testConnectivity):
            """Create a :class:`.TestConnectivityHandler`."""
            super(TestConnectivity.TestConnectivityHandler, self).__init__()
            self.testConnectivity = testConnectivity

        def handleRequest(self, request):
            """
            Handle a test connectivity :class:`.RequestMessage`.

            Support a *GET* and *PUT*
            """
            if request is not None and request.getMethod() is not None:
                method = request.getMethod().upper()
                if method == "GET":
                    return self.getTestConnectivity(request)
                elif method == "PUT":
                    return self.putTestConnectivity(request)
                else:
                    return TestConnectivity.getMethodNotAllowedResponse(request)
            return TestConnectivity.getNullRequestResponse(request)

        def getTestConnectivity(self, request):
            """
            Return the test connectivity parameters.

            :return: a :class:`.RepsonseMessage`
            """
            job = dict()
            job["active"] = False
            if self.testConnectivity.testConnectivityThread is not None:
                with self.testConnectivity.testConnectivityThreadLock:
                    if self.testConnectivity.testConnectivityThread.is_alive():
                        job["active"] = True

            job["count"] = self.testConnectivity.count
            job["currentCount"] = self.testConnectivity.currentCount
            job["interval"] = self.testConnectivity.interval
            job["size"] = self.testConnectivity.size
            return TestConnectivity.getResponseMessage(request,
                                                       job, StatusCode.OK)

        def putTestConnectivity(self, request):
            """
            Start a :class:`.TestConnectivityThread` thread.

            Start a test connectivity thread if a thread is not
            currently active and initialize with the given parameters.
            :param request:
            :return: a :class:`.ResponseMessage`
            """
            # TODO: Is request.getBody() already a json object ?
            jsonObject = json.loads(request.getBody())
            try:
                errors = ""
                active = None
                if "active" in jsonObject:
                    active = jsonObject["active"]
                if active is None:
                    return TestConnectivity.getBadRequestResponse(
                        request, "Required parameter 'active' is missing.")

                if not active:
                    if self.testConnectivity.testConnectivityThread is not None:
                        with self.testConnectivity.testConnectivityThreadLock:
                            if self.testConnectivity.testConnectivityThread.is_alive():
                                self.testConnectivity.testConnectivityThread._stop()
                    return TestConnectivity.getResponseMessage(request, "",
                                                               StatusCode.OK)

                # is active
                if self.testConnectivity.testConnectivityThread is not None:
                    with self.testConnectivity.testConnectivityThreadLock:
                        if self.testConnectivity.testConnectivityThread.is_alive():
                            return TestConnectivity.getAlreadyRunningResponse(request)

                # TODO: Shouldn't the assignements be synchronized ?
                paramtuple = TestConnectivity.getParam(jsonObject, "count",
                                                       errors)
                self.testConnectivity.count = paramtuple[0]
                errors = paramtuple[1]
                if self.testConnectivity.count is None or \
                        self.testConnectivity.count < 1:
                    errors = f'{errors} count must be greater then 0.'

                paramtuple = TestConnectivity.getParam(jsonObject, "interval",
                                                       errors)
                self.testConnectivity.interval = paramtuple[0]
                errors = paramtuple[1]
                if self.testConnectivity.interval is None or \
                        self.testConnectivity.interval < 0:
                    errors = \
                        f'{errors} interval must be greater then or equal to 0.'

                paramtuple = TestConnectivity.getParam(jsonObject, "size",
                                                       errors)
                self.testConnectivity.size = paramtuple[0]
                errors = paramtuple[1]
                if self.testConnectivity.size is None or \
                        self.testConnectivity.size > 4096:
                    errors = f'{errors} size is greater than 4096.'

                if len(errors) > 0:
                    TestConnectivity.resetParameters()
                    return TestConnectivity.getBadRequestResponse(request,
                                                                  errors)

                self.testConnectivity.testConnectivityThread = \
                    TestConnectivity.TestConnectivityThread(
                        self.testConnectivity)
                self.testConnectivity.testConnectivityThread.start()

            except Exception as e:
                # TODO: What is the exception ?
                # TODO: IF there is an exception just return bad request
                # if isinstance(e, (ClassCastException, )) or \
                #         isinstance(e, (JSONException, )) or \
                #         isinstance(e, (NullPointerException, )):
                #     return TestConnectivity.getBadRequestResponse(request,
                #         "Error parsing JSON request, check parameters.")
                # raise e
                return TestConnectivity.getBadRequestResponse(
                    request, "Error parsing JSON request, check parameters.")

            return TestConnectivity.getResponseMessage(request, "",
                                                       StatusCode.OK)

    class TestConnectivityThread(threading.Thread):
        """A `Thread` to test connectivity with IoT server."""

        __slots__ = (
            'payload',
            'testConnectivity',
            'testConnectivityMessage',
            'running'
        )

        def __init__(self, testConnectivity):
            """Test the IoT server connevtivity."""
            super(TestConnectivity.TestConnectivityThread, self).__init__(
                name="iotcs.client.impl.TestConnectivityThread",
                daemon=True)
            self.testConnectivity = testConnectivity
            # TODO: Definitely don't want INTMAX length array.
            sizeInt = 4096 if self.testConnectivity.size > 4096 else \
                self.testConnectivity.size
            self.payload = [0] * sizeInt

            try:
                self.testConnectivityMessage = DataMessage(
                    source=self.testConnectivity.endpointId,
                    format=f'{Diagnostics.DIAGNOSTICS_URN}:test_message')
                # TODO: can a dataItem value be an list ?
                self.testConnectivityMessage.addItem(
                    "count", self.testConnectivity.count)
                # Use json to "stringify" the payload list.
                self.testConnectivityMessage.addItem("payload",
                                                     json.dumps(self.payload))
            except Exception as e:
                raise e
            self.running = False

        # TODO: Use timer not sleep
        def run(self):
            """Test connectivity with the IoT server."""
            self.running = True
            try:
                c = 0
                while self.running and c < self.testConnectivity.count:
                    try:
                        self.testConnectivityMessage.addItem("current", c)
                        self.testConnectivity.messageDispatcher.queue(
                            [self.testConnectivityMessage])
                        self.testConnectivity.currentCount = c + 1
                    except Exception as e:
                        self.getLogger().warning(
                            f'run(): caught exception {e}')
                        pass
                    # units is millisecods
                    time.sleep(self.testConnectivity.interval / 1000)
                    c += 1

            except Exception as ie:
                self.testConnectivity.resetParameters()
            # set the thead instance to None
            with self.testConnectivity.testConnectivityThreadLock:
                self.testConnectivity.testConnectivityThread = None
            return

        def _stop(self):
            """Stop the thread."""
            self.running = False

        def getLogger(self):
            """
            Returen a logger for :class:'.TestConnectivityThread`.

            :return: the looger for :class:`.TestConnectivityThread`
            """
            return logging.getLogger(
                'iotcs.client.impl.TestConnectivity.TestConnectivityThread')
