# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import base64
from http.client import RemoteDisconnected
import json
import logging
import ssl
import threading
import time
from urllib.parse import urlencode
import urllib3
import urllib3.util

#from urllib3.exceptions import ConnectTimeoutError
from urllib3.exceptions import NewConnectionError
from urllib3.exceptions import MaxRetryError
# from urllib3.exceptions import InsecurePlatformWarning
# from urllib3.exceptions import InsecureRequestWarning
# from urllib3.exceptions import ReadTimeoutError
from urllib3.exceptions import SSLError
from urllib3.exceptions import HTTPError
from urllib3.exceptions import HTTPWarning
from urllib3.exceptions import SecurityWarning
from urllib3.exceptions import TimeoutError

from iotcs.common import TimeManager
from iotcs.common import RestApi
from iotcs.common import StatusCode
from iotcs.common import Util
#from iotcs.exception import HttpException
from iotcs.exception import SecurityException
from iotcs.exception import SocketException
from iotcs.exception import SocketTimeoutException
from iotcs.exception import TransportException
from iotcs.transport import AccessToken
from iotcs.transport import SecureConnection
from iotcs.transport import SendReceiveImpl
from iotcs.trust import UnifiedTrustedAssetsManager


class HttpCredentials():

    __slots__ = (
        'tam'
    )
    #15 minutes timeout for renewAccessToken
    EXP_CLAIM_DELTA = 15 * 1000 * 60

    def __init__(self, tam: UnifiedTrustedAssetsManager):
        self.tam = tam

    def getAssertionCredentialsPostData(self):
        useSharedSecret = (self.tam.getEndpointId() is None)
        if (useSharedSecret is True):
            scope = "oracle/iot/activation"
        else:
            scope = ""
        postData = "grant_type=client_credentials"
        postData += "&client_assertion_type=urn%3Aietf%3Aparams%3Aoauth%3Aclient-assertion-type%3Ajwt-bearer"
        postData += "&client_assertion=" + self.buildClientAssertion(useSharedSecret)
        postData += "&scope=" + scope
        return bytearray(postData,"utf-8")

    def buildClientAssertion(self, useSharedSecret):
        millis = int(round(time.time() * 1000))
        val = int((millis + HttpCredentials.EXP_CLAIM_DELTA) / 1000)
        #exp = numpy.uint32(val)
        #exp = ctypes.c_uint32(val)
        if (useSharedSecret):
            id = self.tam.getClientId()
        else:
            id = self.tam.getEndpointId()
        if (useSharedSecret):
            alg = "HS256"
        else:
            alg = "RS256"
        header_dict = {
            'typ': 'JWT',
            'alg': alg,
        }
        header = '{"typ":"JWT","alg":"'+ alg + '"}'
        #print(header)
        claims = '{"iss":"' + id + '", "sub":"' + id + '", "aud":"' + 'oracle/iot/oauth2/token' + '", '
        claims += '"exp":' + str(val) + '}'
        #print(claims)
        try:
            inputToSign = str(base64.urlsafe_b64encode(header.encode()),'utf-8')
            inputToSign += "."
            inputToSign += str(base64.urlsafe_b64encode(claims.encode()),'utf-8')
            inputdata = bytearray(inputToSign,'utf-8')
            if (useSharedSecret):
                signedData=self.tam.signWithSharedSecret(inputdata)
            else:
                signedData=self.tam.signWithPrivateKey(inputdata)
            signature = str(base64.urlsafe_b64encode(signedData),'utf-8')
            inputToSign += "."
            inputToSign += signature
            return inputToSign
        except Exception as e:
            print(e)
            raise e


class HttpSecureConnection(SecureConnection):
    """
    - messaging:long_polling_timeout_offset :	Default: 100 milliseconds
        The time, in milliseconds, added to the receive timeout given by the
        application and then used for the client transport timeout.
    - messaging:send_receive_timeout :	Default: 100 milliseconds
        Only used when long polling is disabled. The minimum time, in
        milliseconds, since the last send before polling the server for
        incoming messages if the receive buffer is empty. A larger
        value may cause a delay in the receipt of incoming messages, but will
        cause less network overhead. A value of zero will cause the library to
        skip the poll for incoming messages when calling receive
        if the receive buffer is empty.
    - messaging:http_response_timeout :	Default: 15000 milliseconds
        Only used when long polling is disabled. The amount of time, in
        milliseconds, to wait for the server to respond to an HTTPS request.

    When long polling the server closes the connection after 9 minutes.
    """

    __slots__ = (
        # TODO: This is defined in SecureConnection use
        # TODO: getTrustedAssetsManater
        '__tam',
        '__server',
        '__port',
        '__scheme',
        '__credentials',
        '__baseServerUrl',
        '__currentAccessToken',
        '__http',
        '__tokenlock',
        '__defaultHeaders',
        '__maxRetries'
    )

    OAUTH_ENDPOINT = "/iot/api/v2/oauth2/token"
    ACCESS_TOKEN_MAX_RETRIES_PROPERTY = "access_token_max_retries"

    def __init__(self, tam: UnifiedTrustedAssetsManager):
        super(HttpSecureConnection, self).__init__(tam)

        ssl._create_default_https_context = ssl._create_unverified_context

        self.__tam = tam
        self.__server = tam.getServerHost()
        self.__port = tam.getServerPort()
        self.__scheme = tam.getServerScheme()
        self.__credentials = HttpCredentials(tam)
        self.__baseServerUrl = self.__scheme + "://" + self.__server + ":" + \
                               str(self.__port)
        self.__currentAccessToken = None
        context = ssl.SSLContext()
        # TODO: Why does CERT_NONE appear unresolved ? docs claim it is defined
        # TODO: in ssl
        context.verify_mode = ssl.CERT_NONE

        # TODO: encapsulate http library.
        timeout = Util.getConfig().getint("http",
                                          "http_connection_timeout")
        if timeout is None or timeout == -1:
            self.__http = urllib3.PoolManager(retries=0,
                                          cert_reqs="CERT_NONE")
        else:
            self.__http = urllib3.PoolManager(retries=0,
                                              timeout=timeout,
                                              cert_reqs="CERT_NONE")

        urllib3.disable_warnings()
        self.__tokenlock = threading.Lock()

        self.__maxRetries = Util.getConfig().getint(
            Util.CONFIG_DEVICE_MESSAGING_SECTION,
            HttpSecureConnection.ACCESS_TOKEN_MAX_RETRIES_PROPERTY)

        # TODO: These can be added to the coonection pool to always
        # TODO: include these headers in all requests
        self.__defaultHeaders = dict()
        self.__defaultHeaders["Accept"] = "application/json"
        self.__defaultHeaders["Accept-Encoding"]= "deflate"
        self.__defaultHeaders["Content-Type"] = "application/json"

    def getResolvedUrl(self, endpoint, queryParams):
        """

        :param endpoint:
        :param queryParams:
        :return:
        """
        resolvedUrl = self.__baseServerUrl + endpoint
        if (queryParams is not  None):
            urlparams = urlencode(queryParams)
            resolvedUrl = resolvedUrl + "?" + urlparams
            #print(resolvedUrl)
        return resolvedUrl

    def invalidateAccessToken(self):
        with self.__tokenlock:
            self.__currentAccessToken = None

    def postRenewAccessToken(self):
        """
        """

        # TODO: All http requests must eventually go through invoke
        # TODO: Must use self.post, which must accept headers.

        credentials = self.__credentials.getAssertionCredentialsPostData()

        headers=dict()
        headers["Content-Type"] = "application/x-www-form-urlencoded"
        headers["Accept"] = "application/json"

        url = self.getResolvedUrl(HttpSecureConnection.OAUTH_ENDPOINT, None)
        # resp = self.__http.request("POST", url, fields=None, headers=headers,
        #                            body = credentials)
        # status = resp.status

        # Should raise exception or return non None HttpResponse
        # use invoke directly since adding access token headers occurs in
        # higher level methods, i.e. get, post, ...
        httpresponse = self._invoke(url, "POST", body=credentials,
                                    headers=headers)
        return httpresponse

    def renewAccessToken(self, retries=-1):
        """

        :return:
        :raisees: Exceptions raised by urllib3.PoolManager.request.
        """

        # # JCL "postRenewAccessToken"
        #
        # # TODO: All http requests must eventually go through invoke
        # # TODO: Must use self.post, which must accept headers.
        #
        # credentials = self.__credentials.getAssertionCredentialsPostData()
        #
        # headers=dict()
        # headers["Content-Type"] = "application/x-www-form-urlencoded"
        # headers["Accept"] = "application/json"
        #
        # url = self.getResolvedUrl(HttpSecureConnection.OAUTH_ENDPOINT, None)
        # # resp = self.__http.request("POST", url, fields=None, headers=headers,
        # #                            body = credentials)
        # # status = resp.status
        #
        # # Should raise exception or return non None HttpResponse
        # # use invoke directly since adding access token headers occurs in
        # # higher level methods, i.e. get, post, ...
        # httpresponse = self._invoke(url, "POST", body=credentials,
        #                             headers=headers)

        httpresponse = self.postRenewAccessToken()

        # JCL "renewAccessToken"

        status = httpresponse.getStatus()
        if status == 400:
            respdata = httpresponse.getData()
            resptext = str(respdata,'utf-8')
            # If the request fails there won't be "currentTime" in the response
            # return None and let the caller handle it
            ind = None
            try:
                ind = resptext.index("currentTime")
                if ind > 0:
                    respJson = json.loads(resptext)
                    currentTimeStr = respJson("currentTime")
                    currentTime = int(currentTimeStr)
                    TimeManager.setCurrentTimeMillis(currentTime)
                    httpresponse = self.postRenewAccessToken()
                    status = httpresponse.getStatus()
            except Exception as e:
                self.getLogger().warning(
                    f'renewAccessToken(): '
                    f'status {httpresponse.getStatus()}')
                #return None

        if status == 400:
            raise SecurityException(httpresponse.getVerboseStatus("POST",
                    HttpSecureConnection.OAUTH_ENDPOINT))

        if status != 200:
            raise TransportException(status, HttpSecureConnection.OAUTH_ENDPOINT)

        data = httpresponse.getData()
        if data is None or len(data) == 0:
            self.getLogger().warning(
                f'HttpSecureConnection.renewAccessToken(): '
                f'httpresponse.getData() returned None or empty')
            raise IOError(f'HttpSecureConnection.renewAccessToken(): '
                          f'httpresponse.getData() returned None or empty')
            #return None

        # TODO: doe not need to be converted to str
        #jsondata = str(data,'utf-8')
        return AccessToken.fromJson(data)

        # TODO: REMOVE
        # # 401 means the credentials have expired
        # # 403 means we're probably using client-secret where we need
        # # client-credentials
        # if httpresponse.getStatus() == 403 or httpresponse.getStatus() == 401:
        #     respdata = httpresponse.getData()
        #     resptext = str(respdata,'utf-8')
        #     # If the request fails there won't be "currentTime" in the response
        #     # return None and let the caller handle it
        #     ind = None
        #     try:
        #         ind = resptext.index("currentTime")
        #     except Exception as e:
        #         self.getLogger().warning(
        #             f'renewAccessToken(): '
        #             f'status {httpresponse.getStatus()}')
        #         return None
        #
        #     # TODO: implement retry count
        #     # Looks like this could end up in an infinite loop
        #     if ind > 0:
        #         respJson = json.loads(resptext)
        #         currentTimeStr = respJson("currentTime")
        #         currentTime = int(currentTimeStr)
        #         TimeManager.setCurrentTimeMillis(currentTime)
        #
        #         if retries == -1:
        #             retries = self.__maxRetries
        #         else:
        #             retries -= 1
        #
        #         if retries > 0:
        #             return self.renewAccessToken(retries=retries)
        #         else:
        #             self.getLogger().warning(
        #                 f'renewAccessToken(): '
        #                 f'maximum retries {self.__maxRetries} exceeded')
        #
        #     return None

        # TODO: REMOVE
        # data = httpresponse.getData()
        # if data is None:
        #     self.getLogger().warning(
        #         f'HttpSecureConnection.renewAccessToken(): '
        #         f'httpresponse.getData() returned None')
        #     return None
        #
        # # TODO: doe not need to be converted to str
        # #jsondata = str(data,'utf-8')
        # return AccessToken.fromJson(data)

    # TODO: Non-reentrant
    def getAccessToken(self, force=False):
        token = None
        with self.__tokenlock:
            if force or self.__currentAccessToken is None or \
                    self.__currentAccessToken.hasExpired():
                # TODO: Should have thrown TransportException with status
                accessToken = self.renewAccessToken()
                if accessToken is None:
                    raise RuntimeError(f'Could not get OAUTH accessToken for '
                        f'{self.__tam.getClientId()}')
            else:
                accessToken = self.__currentAccessToken

            self.__currentAccessToken = accessToken
            token = self.__currentAccessToken.getTokenType() + " " + \
                        self.__currentAccessToken.getToken()

        return token


    def addAccessToken(self, headers):

        # TODO: What about "isWebApi" ?
        # TODO: or testing for "isActivationApi" ?

        # Add these headers always for "activation api's"
        # and for all requests, unless it "isWebApi"
        if self.__tam.isActivated():
            headers['X-EndpointId'] = self.__tam.getEndpointId()
        else:
            headers['X-ActivationId'] = self.__tam.getClientId()

        # TODO: Do not add access token header if "isWebApi"
        access_token = self.getAccessToken()
        headers['Authorization'] = access_token

    def disconnect(self):
        """Disconnect the connection to the IoT server."""
        self.invalidateAccessToken()

    def get(self, url, queryparams, headers=None, timeout=None):
        """
        Http GET
        :param url: REST resource
        :param queryparams: HTTP query parameters
        :param headers: these headers are merged with required headers and can
            override a required header
        :return: response as `iotcs.shared.http.HttpResponse`
        """

        _headers = dict(self.__defaultHeaders)
        if headers is not None:
            _headers.update(headers)
        self.addAccessToken(_headers)
        resolvedUrl = self.getResolvedUrl(url, queryparams)
        return self._invoke(resolvedUrl, "GET", headers=_headers,
                            timeout=timeout)

    def post(self, url, queryparams, data, headers=None, timeout=None):
        """

        :param url:
        :param queryparams:
        :param data:
        :param headers:
        :param timeout:
        :return: `iotcs.shared.http.HttpResponse`
        """

        _headers = dict(self.__defaultHeaders)
        if headers is not None:
            _headers.update(headers)
        self.addAccessToken(_headers)
        resolvedUrl = self.getResolvedUrl(url, queryparams)

        # TODO: If it is not str what is the expectation, byte-like ?
        # Should be json object (python object) now.
        if isinstance(data, str):
            print(f'\n### WARNING - HttpSecureConnection.post(): '
                  f'expecting object got str: {str} ###\n')
            data = bytearray(data,'utf-8')

        return self._invoke(resolvedUrl, "POST", body=data, headers=_headers,
                            timeout=timeout)

    def put(self, url, queryparams, data, headers=None, timeout=None):
        _headers = dict(self.__defaultHeaders)
        if headers is not None:
            _headers.update(headers)
        self.addAccessToken(_headers)
        resolvedUrl = self.getResolvedUrl(url, queryparams)
        return self._invoke(resolvedUrl, "PUT", body=data, headers=_headers,
                            timeout=timeout)

    def head(self, url, queryparams, data, headers=None, timeout=None):
        _headers = dict(self.__defaultHeaders)
        if headers is not None:
            _headers.update(headers)
        self.addAccessToken(_headers)
        resolvedUrl = self.getResolvedUrl(url, queryparams)
        return self._invoke(resolvedUrl, "HEAD", body=data, headers=_headers,
                            timeout=timeout)

    def delete(self, url, queryparams, data, headers=None, timeout=None):
        _headers = dict(self.__defaultHeaders)
        if headers is not None:
            _headers.update(headers)
        self.addAccessToken(_headers)
        resolvedUrl = self.getResolvedUrl(url, queryparams)
        return self._invoke(resolvedUrl, "DELETE", body=data, headers=_headers,
                            timeout=timeout)

    def patch(self, url, queryparams, data, headers=None, timeout=None):
        _headers = dict(self.__defaultHeaders)
        if headers is not None:
            _headers.update(headers)
        _headers["X-HTTP-Method-Override"] =  "PATCH"
        self.addAccessToken(_headers)
        resolvedUrl = self.getResolvedUrl(url, queryparams)
        return self._invoke(resolvedUrl, "POST", body=data, headers=_headers,
                            timeout=timeout)

    # HTTPError Base for all ulrlib exceptions
    # We just pass these on.
    # We treat SSLErrors like Socket errors
    # ConnectTimeOutError Raised when a socket timeout occurs while
    # connecting to a server
    # ReadTimeoutError Raised when a socket timeout occurs while
    # receiving data from a server or
    # Raised when a socket timeout error occurs.
    # Catching TimeoutError error will catch both ReadTimeoutErrors
    # and ConnectTimeoutErrors.
    # SSLError, Raised when SSL certificate fails in an HTTPS connection.
    _sockettimeoutexceptions = (TimeoutError,SSLError,)
    # SecurityWarning Warned when performing security reducing actions
    # or Warned when performing security reducing actions
    _securityexceptions = (SecurityWarning,)
    # Urllib3 implements unexpected behavior with retries.
    # raises MaxRetryError when retries is disabled
    # raises.
    # NewConnectionError is like UnknownHostException, or unreachable
    _socketexceptions =  (NewConnectionError,MaxRetryError,)
    # RemoteDisconnected Raised when connection just closes on its own.
    # From reading its not clear if this is a urllib3 problem.
    # We see this from mock server after every request when it calls
    # HttpExchange.close(). Have not seen it from IOT server.
    # (still need to verify this)
    _transportexceptions = (RemoteDisconnected,)

    def _invoke(self, url, meth, body=None, headers=None, timeout=None):
        """

        :param url:
        :param meth:
        :param reqdata: json object python object
        :param headers:
        :param timeout:
        :return: response as `iotcs.shared.http.HttpResponse
        """

        payload = body
        if body is not None and not isinstance(body, (bytes,bytearray)):
            body = json.dumps(body)
            body = bytearray(body,'utf-8')
        # elif body is not None:
        #     # TODO: Remove - This case is so far only renewAccessToken
        #     print(f'\n### WARNING - HttpSecureConnection._invoke(): expecting
        #           object got bytes or bytearray: {str(body)} ###\n')

        if timeout is not None:
            _timeout = urllib3.util.Timeout()

        # request expects "byte-like" object
        # response is 'urllib3.response.HTTPResponse'
        response = None
        try:
            # TODO: Implement retries for access token
            if body is not None:
                response = self.__http.request(meth, url, fields=None, retries=0,
                                               headers=headers, body=body)
            else:
                response = self.__http.request(meth, url, fields=None, retries=0,
                                               headers=headers)

            # 401 means the credentials have expired
            # 403 means we're probably using client-secret where we need
            # client-credentials
            if response.status == 401 or response.status == 403:
                access_token = self.getAccessToken(force=True)
                headers['Authorization'] = access_token
                if body is not None:
                    response = self.__http.request(meth, url, fields=None, retries=0,
                                                   headers=headers, body=body)
                else:
                    response = self.__http.request(meth, url, fields=None, retries=0,
                                               headers=headers)

        # Seeing this if the attemtpt to connect fails, say no VPN to iot server
        except HttpSecureConnection._socketexceptions as socketexception:
            logging.debug(f'HttpSecureConnection._invoke(): {socketexception}')
            raise SocketException(socketexception)
        except HttpSecureConnection._sockettimeoutexceptions as sockettimeoutexception:
            logging.debug(f'HttpSecureConnection._invoke(): {sockettimeoutexception}')
            raise SocketTimeoutException(sockettimeoutexception)
        except HttpSecureConnection._securityexceptions as securityexception:
            logging.debug(f'HttpSecureConnection._invoke(): {securityexception}')
            raise SecurityException(securityexception)
        except HttpSecureConnection._transportexceptions as transportexception:
            logging.debug(f'HttpSecureConnection._invoke(): {transportexception}')
            raise TransportException(StatusCode.SERVICE_UNAVAILABLE.value, f'{url}')
        # TODO: Not sure if these should be special cases
        # except (HTTPError,HTTPWarning,) as httpexception:
        #     raise HttpException(httpexception)

        if response.status == 400:
            logging.critical(f'HttpSecureConnection._invoke(): '
                f'{response.getVerboseStatus(meth, url)}')

        # TODO: PUSH UP
        # Check response for 503 and throw TransportException
        if response.status == StatusCode.SERVICE_UNAVAILABLE:
            raise TransportException(StatusCode.SERVICE_UNAVAILABLE.value, f'{url}')

        return HttpResponse(response=response)

    def close(self):
        super(HttpSecureConnection, self).close()
        # Does not close in flight connections
        try:
            self.__http.clear()
        except:
            pass

    def getLogger(self):
        return logging.getLogger('iotcs.device.impl.HttpSecureConnection')


MESSAGE_ENDPOINT=f'{RestApi.V2.getReqRoot()}/messages'

class HttpSendReceiveImpl(SendReceiveImpl):
    """
    `HttpSendReceiveImpl` is an implementation of the `DirectlyConnectedDevice.send()`.
    The send is synchronous. Receive should be considered
    as a synchronous call. The implementation has a buffer for receiving request messages.
    If there are no messages in the buffer, the receive implementation will send a message
    to the server to receive any pending requests the server may have.

    The size of the buffer can be configured see `iotcs.device.impl.SendReceiveImpl`
    """

    __slots__ = (
        'longPollingOffset',
        'secureConnection'
    )

    MIN_ACCEPT_BYTES_HEADER = "X-min-acceptBytes"

    def __init__(self, secureConnection):
        super(HttpSendReceiveImpl, self).__init__()
        self.secureConnection = secureConnection
        self.longPollingOffset = Util.getConfig().getint(Util.CONFIG_DEVICE_MESSAGING_SECTION,
                                        SendReceiveImpl.LONG_POLLING_TIMEOUT_OFFSET_PROPERTY)

    # TODO: see SendReceiveImpl. Called from SendRecieveImpl and MQTTSendReceiveImpl receive
    # payload is json object (python object)
    def post(self, payload, timeout=SendReceiveImpl.USE_DEFAULT_TIMEOUT_VALUE):
        """
        HTTP 'POST' payload to IOT Cloud Service
        :param payload: json object (pyhton object) could be `list` or `dict`
        :param timeout:
        :return:
        """

        availableBytes = 0
        if not self.useLongPolling or payload is None:
            # head and tail are kept so that they are less than requestBuffer.length.
            # This avoids the (unlikely) situation that they would overflow.
            # But this also means that tail might be less than head, since the buffer
            # is circular. This has to be taken into account when calculating the
            # available bytes. Remember that tail is the pointer to the next
            # available chunk and does not count toward used bytes (no need to
            # adjust used byte count by 1). If tail == head, then used bytes is
            # zero - perfect!

            # PCL: usedBytes are accounted for in "getAvailableBytes"

            # subtract 2 for the length bytes.
            # TODO: subtracting 2 may fail if there is more than one message
            # in the received buffer. Say we had 100 bytes available and we two
            # messages of 49 bytes. We'd write 2 bytes for the 49, then the
            # 49 bytes. Now we only have 49 bytes of buffer left, which is not
            # enough to hold 2 bytes of length plus 49 bytes of message.

            # PCL: accounted for in getAvailableBytes()
            availableBytes = self.requestBuffer.getAvailableBytes()

        # TODO: Why doesn't this and other parameters use "query params" object ? (dict)
        restApi = f'{MESSAGE_ENDPOINT}?acceptBytes={availableBytes}'

        longpolling = self.useLongPolling and availableBytes > 0
        if longpolling:
            restApi = f'{restApi}&iot.sync'
            if timeout > 0:
                #iot.timeout is in seconds
                iotTimeout = timeout / 1000
                if iotTimeout == 0:
                    iotTimeout = 1

                restApi = f'{restApi}&iot.timeout={iotTimeout}'
                # Add an offset to the transport level timeout
                # as a failover mechanism, just in case.
                timeout = (iotTimeout * 1000) + self.longPollingOffset

        # FINER
        if self.getLogger().isEnabledFor(logging.DEBUG):
            self.getLogger().debug(f'post(): {restApi} : {json.dumps(payload)}')

        response = None
        try:
            response = self.secureConnection.post(restApi, None, payload, timeout=timeout)
        except SocketTimeoutException as ste:
            # Do not throw the SocketTimeoutException if the post was for long polling.
            # The SocketTimeoutException is expected for long polling and just
            # means that the server closed its end.
            # TODO: This might be RemoteDisconnected, if the server does not close the
            # TODO: connection properly
            if longpolling:
                if self.getLogger().isEnabledFor(logging.DEBUG):
                    self.getLogger().debug(f'post(): caught SocketTimeoutException {ste} for {restApi}')
                return
            else:
                raise ste
        except Exception as e:
            # TODO: Do we need more specialization ?
            raise e

        # FINEST
        if self.getLogger().isEnabledFor(logging.DEBUG):
            self.getLogger().debug(f'post(): {response.getVerboseStatus("POST", restApi)}')

        status = response.getStatus()
        if status == StatusCode.OK or status == StatusCode.ACCEPTED:
            if availableBytes == 0:
                # TODO: What does this mean, request messages are thrown away ?
                return

            headers = response.getHeaders()
            header = None
            if SendReceiveImpl.MIN_ACCEPT_BYTES_HEADER in headers:
                header = headers[SendReceiveImpl.MIN_ACCEPT_BYTES_HEADER]
                if header is not None and len(header) > 0:
                    minAcceptBytes = int(header[0])
                    # The next request the server has for this client is larger
                    # than the bytes available in the request buffer.
                    #
                    # A properly configured and responsive system should never
                    # get in this state.
                    # TODO: This should be in RequestBuffer
                    buffersize = self.requestBuffer.getSize()
                    usedBytes = self.requestBuffer.getUsedBytes()
                    if minAcceptBytes > buffersize - 2:
                        # SEVERE
                        self.getLogger().error(f'The server has a request of {minAcceptBytes} bytes '
                            f'for this client which is too large for the {buffersize} byte '
                            f'request buffer. Please restart the client with a larger value '
                            f'for the {Util.CONFIG_DEVICE_MESSAGING_SECTION}:'
                            f'{SendReceiveImpl.REQUEST_BUFFER_SIZE_PROPERTY} '
                            f'in the "iotcs.config.config.ini" file')
                    else:
                        # The message(s) from the last time have not been processed
                        self.getLogger().warning(f'The server has a request of {minAcceptBytes} '
                            f'for this client, which cannot be sent because the {buffersize} byte '
                            f'request buffer is filled with {usedBytes} of unprocessed requests')
                    return

            if response.getData() != b'[]':
                self.getLogger().debug(f'post(): {response.getData()}')

            self.requestBuffer.write(response.getData())
        else:
            if longpolling and status == StatusCode.HTTP_GATEWAY_TIMEOUT:
                # Do not throw the exception when long polling.
                if self.getLogger().isEnabledFor(logging.DEBUG):
                    self.getLogger().debug(f'post(): {response.getVerboseStatus("POST", restApi)}')
                return
            raise TransportException(status, response.getVerboseStatus('POST', restApi))

    def getLogger(self):
        return logging.getLogger('iotcs.device.impl.HttpSendReceiveImpl')

class HttpResponse():
    """
    HttpResponse
    """
    __slots__ = (
        '__status',
        '__data',
        '__headers'
    )

    def __init__(self, responseCode=None, data=None, headers=None, response=None):
        """

        :param responseCode:
        :param data: b'', bytearray
        :param headers:
        :param response: if not None must be instance of `urllib3.response.HTTPResponse`
        """
        # First check for response, and if other parameters are not None, override the
        # values from response
        if response is not None:
            # TODO: convert to StatusCode
            self.__status = response.status
            self.__data = response.data
            self.__headers = response.getheaders()

        # int
        if responseCode is not None:
            self.__status = responseCode

        # b''
        if data is not None:
            self.__data = data

        # urllib3._collections.HTTPHeaderDict
        if headers is not None:
            self.__headers = headers

    def getStatus(self):
        return self.__status

    def getData(self):
        return self.__data

    def getHeaders(self):
        return self.__headers

    def getVerboseStatus(self, method, path):

        verboseStatus = ""
        if method is not None:
            verboseStatus = f'{verboseStatus}{method} '
        
        if path is not None:
            verboseStatus = f'{verboseStatus}{path}: '

        verboseStatus = f'{verboseStatus}HTTP {self.__status}'

        if self.__status < StatusCode.OK or self.__status >= StatusCode.BAD_REQUEST:
            if self.__data is not None and len(self.__data) > 0:
                verboseStatus = f'{verboseStatus} {str(self.__data)}'

        return verboseStatus

    def __str__(self):
        return self.getVerboseStatus(None, None)

