# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import logging

from iotcs.client.model import DeviceModelFactory
from iotcs.activation import DIRECT_ACTIVATION_URN
from iotcs.activation import INDIRECT_ACTIVATION_URN
from iotcs.activation import createIndirectActivationRequest
from iotcs.activation import createDirectActivationRequest
from iotcs.activation import getActivationPolicy
from iotcs.activation import postIndirectActivationRequest
from iotcs.activation import postDirectActivationRequest
from iotcs.exception import SecurityException
from iotcs.exception import StateException
from iotcs.http import HttpSecureConnection
from iotcs.http import HttpSendReceiveImpl
from iotcs.message import AlertMessage
from iotcs.message import DataMessage
from iotcs.message import DataItem
from iotcs.message import ResourceMessage
from iotcs.message import Resource
from iotcs.messaging.client.device.policy import DevicePolicyManager
from iotcs.messaging.client.device.policy import MessagingPolicy
from iotcs.messaging.client.device.storage import StorageConnectionImpl
from iotcs.persistence import PersistenceStoreManager
from iotcs.trust import UnifiedTrustedAssetsManager
from iotcs.common import Closeable
from iotcs.common import TimeManager

MESSAGE_QUEUE_SIZE = 100

# com.oracle.iot.client.device

# TODO: why is this map needed ?
# TODO: If there is a one to one relationship between a SecureConnection
# TODO: and a StorageConection and a one to one relationship between a
# TODO: dcd and a SecureConnection, why isn't this just a field ?

_STORAGE_CONNECTION_MAP = dict()

class DirectlyConnectedDeviceImpl(Closeable):

    @classmethod
    def getStorageConnection(cls, secureConnection):
        """
        Return the :class:`.StorageConnection` with this `secureConnection`.

        :param secureConnection: a :class:`.SecureConnection` instance
        :return: return the :class:`.StorageConnection` mapped to
            `secureConnection`
        """
        storageConnection = None
        if secureConnection in _STORAGE_CONNECTION_MAP:
            storageConnection = _STORAGE_CONNECTION_MAP[secureConnection]
        if storageConnection is None:
            storageConnection = StorageConnectionImpl(secureConnection)
            _STORAGE_CONNECTION_MAP[secureConnection] = storageConnection
        return storageConnection

    @classmethod
    def closeStorageConnection(cls, secureConnection):
        if secureConnection in _STORAGE_CONNECTION_MAP:
            storageConnection = _STORAGE_CONNECTION_MAP.pop(secureConnection)
            storageConnection.close()

    USE_DEFAULT_TIMEOUT_VALUE = -1
    DIRECTLY_CONNECTED_DEVICE_CAPABILITIES = ''

    __slots__ = (
        'tam',
        'scheme',
        'isMqtt',
        'secureConn',
        'sendReceiveImpl',
        # To overcome circular dependendency
        'isGatewayDevice'
    )

    def __init__(self, path, passphrase, isGatewayDevice=False):
        """
        This constructor is used internally and not intended for general use.
        :param path:
        :param passphrase:
        :param isGatewayDevice:
        """
        super(DirectlyConnectedDeviceImpl, self).__init__()
        self.isGatewayDevice = isGatewayDevice
        self.tam = UnifiedTrustedAssetsManager(path, passphrase)
        self.scheme = self.tam.getServerScheme()
        self.isMqtt = (self.scheme == "mqtts")
        # TODO: Need to verify if "http" is really needed for testing.
        if not self.isMqtt and self.scheme != "https":
                #and \
                #self.scheme != "http"): # need http for testing
            raise SecurityException("DirectlyConnectedDeviceImpl.__init__(): Unsupported server scheme")

        self.secureConn = self.createSecureConnection()

        if self.tam.isActivated():
            self.sendReceiveImpl = self.createSendReceiveImpl()
            persistenceStore = PersistenceStoreManager.getPersistenceStore(
                self.getEndpointId())
            persistenceStore.openTransaction()\
                .put(DevicePolicyManager.__name__,
                     DevicePolicyManager(self.secureConn))\
                .commit()
        else:
            self.sendReceiveImpl = None

    def isActivated(self):
        return self.tam.isActivated()

    def getDeviceModel(self, urn):
        dm = DeviceModelFactory.getDeviceModel(self.secureConn, urn)
        return dm

    def getEndpointId(self):
        return self.tam.getEndpointId()

    def createStorageObject(self, uri, name=None, contentType=None, mutable=True):
        """
        Create a new `StorageObject`.

        if `uri` is not `None`, `name` and `contentType` are ignored...
            Create a new :class:`.StorageObject` from the `uri` for a named object
            in storage.

        if `uri` is `None`...
            The created storage object that will have a name with the given
            object name prefixed with the device's endpoint ID and a directory
            separator. The prefix addition can be disabled by setting the
            :mod:`.config` property *disable_storage_object_prefix* in section
            *media_storage* to `True`.

            if `contentType` is `None`, the mime-type defaults to
            *application/octet-stream*.

        :param uri: the URI of the object in the storage cloud, if not
            `None`, `name` and `contentType` are ignored.
        :param name: the unique name to be used to reference the content in
                     storage
        :param contentType: The mime-type of the content or `None`
        :param mutable: If `mutable` is `True` the storage object is  mutable.
            If `mutable` is `False`, then the :class:`.StorageObject`
            will be made immutable by setting it’s *Write-Once-Read-Many*
            (WORM) policy when uploading it to the container to prevent the
            users from deleting it for a 7 days and modifying it.
            Number of days is 7 by default and can be overridden by setting
            property :module:`.config` *storage_worm_days*. If the
            property is *0* then no WORM policy will be applied.
        :return: a :class:`StorageObject <.messaging.client.device.StorageObjectDelegate`  # noqa: E501
        :raises TransportError: if exception raised by the
            runtime, or an abnormal response from the storage cloud
        :raises SecurityException: if there is an exception establishing
            a secure connection to the storage cloud
        """
        # iotcs.messaging.client.StorageConnectionBase
        storageConnection = DirectlyConnectedDeviceImpl.getStorageConnection(
            self.secureConn)

        if uri is not None:
            return storageConnection.createStorageObject(storageUrl=uri)
        else:
            return storageConnection.createStorageObject(
                clientId=self.getEndpointId(), name=name,
                contentType=contentType, mutable=mutable)

    def activate(self, deviceModels):
        """

        :param deviceModels:
        :return:
        :raise ValueError: if device is already activated
        """
        # TODO: Create Exception classes tkat match i.e. State, Argument, etc.
        if self.tam.isActivated():
            raise ValueError(msg = 'Already activated')

        activation_policy = \
            getActivationPolicy(self.secureConn, self.tam.getClientId())
        privateKey, publicKey = \
            self.tam.generateKeyPair(activation_policy.getKeyType(),
                activation_policy.getKeySize())

        deviceModelsSet = set()
        for m in deviceModels:
            deviceModelsSet.add(m)

        deviceModelsSet.add(DIRECT_ACTIVATION_URN)
        if self.isGatewayDevice:
            deviceModelsSet.add(INDIRECT_ACTIVATION_URN)

        direct_activation_request = \
            createDirectActivationRequest(self.tam,
                activation_policy.getHashAlgorithm(), deviceModelsSet)

        # TODO: log FINEST dump direct activation request

        direct_activation_response = \
            postDirectActivationRequest(self.secureConn,
                direct_activation_request)

        # TODO: log FINEST dump direct activation response

        endpointId = direct_activation_response.getEndpointId()
        certificate = direct_activation_response.getCertificate()
        self.tam.setEndpointCredentials(endpointId, certificate)

        # requset to update access token
        self.secureConn.disconnect()

        self.sendReceiveImpl = self.createSendReceiveImpl()

        persistenceStore = PersistenceStoreManager.getPersistenceStore(
            self.getEndpointId())
        persistenceStore.openTransaction() \
            .put(DevicePolicyManager.__name__,
                 DevicePolicyManager(self.secureConn)) \
            .commit()

        self.registerDevicePolicyResource()


    def createSecureConnection(self):
        if not self.isMqtt:
            return HttpSecureConnection(self.tam)
        else:
            #implement MqttSecureConnection
            pass

    def registerDevicePolicyResource(self):

        if not self.isActivated():
            return

        resource = Resource(endpointName=self.getEndpointId(),
                            methods=Resource.Method.PUT,
                            path=DevicePolicyManager.PATH,
                            name=DevicePolicyManager.CAPABILITY_URN)

        resourceMessage = ResourceMessage(endpointName=self.getEndpointId(),
                                          source=self.getEndpointId())
        resourceMessage.register(resource)
        self.sendReceiveImpl.send([resourceMessage])

    def send(self, messages):
        """
        Send messages to the IOT Cloud Service
        :param messages: a `list` of `Message`
        :raises: `ValueError` if device is not activated
        """

        if self.isActivated():
            # FINE
            if self.getLogger().isEnabledFor(logging.DEBUG):
                dumpedmessages = ""
                if len(messages) > 1:
                    self.getLogger().debug(f'send(): '
                        f'endpointId = {self.getEndpointId()} : Send : '
                        f'({len(messages)} messages follow)')
                counter = 1
                nl = ""
                for message in messages:
                    dumpedmessage = self.dumpMessage(message)
                    dumpedmessages = f'{dumpedmessages}{nl}{counter}: {dumpedmessage}'
                    nl = "\n"
                    counter += 1

                logmsg = f'send(): {self.getEndpointId()} : Send : {len(messages)}\n'\
                    f'{dumpedmessages}'
                self.getLogger().debug(logmsg)

            self.sendReceiveImpl.send(messages)
        else:
            raise StateException("DirectlyConnectedDeviceImpl.send(): not activated")

    def offer(self, messages):
        """
        Offer messages to be sent to the server.

        Whether or not the messages are sent depends on policies that have been
        configured for the attributes in the message. If there are no policies
        for the attributes in the message, then this call is equivalent to
        calling {@link #send(Message...)}.
        Depending on the policies, it is possible that all, some, or none of
        the messages will be sent.

        If messages are sent, this call will block until the server
        responds to the message delivery, or a network timeout occurs. The
        network timeout can be configured by setting the :mod:`iotcs.config`
        property *http_response_timeout* in section *device_messaging*.
        :param messages: zero or more :class:`.DataMessage`,
            :class:`.AlertMessage`, :class:`.ResponseMessage` and/or
            :class:`.ResourceMessage` objects.
        :raises: :class:`.TransportException` if there is an I/O exception when
            trying to send messages.
        :raises: :class:`.SecurityException` when key or signature algorithm
            class cannot be loaded, or the key is not in the trusted assets
            store, or the private key is invalid
        :raises: :class:`.StateException` if the gateway device has not been
            authenticated with the server.
        """
        if self.isActivated():
            # We need to distinguish between an empty list of messages
            # that has been passed in versus an empty list of message
            # that has resulted from policies being applied.
            #  So if the list we receive is empty, let send handle it.
            if messages is None or len(messages) == 0:
                self.send(messages)

            messagingPolicy = None
            persistenceStore = \
                PersistenceStoreManager.getPersistenceStore(self.getEndpointId())
            with persistenceStore.getLock():
                messagingPolicy = \
                    persistenceStore.getOpaque(MessagingPolicy.__name__, None)
                if messagingPolicy is None:
                    messagingPolicy = MessagingPolicy(self)
                    persistenceStore.openTransaction()\
                        .putOpaque(MessagingPolicy.__name__, messagingPolicy)\
                        .commit()
                    devicePolicyManager = \
                        DevicePolicyManager.getDevicePolicyManager(self)
                    devicePolicyManager.addChangeListener(messagingPolicy)

            # Now we know here that messages list is not empty.
            # If the message list is not empty after applying policies,
            # then send the messages. If it is empty after applying the
            # policies, then there is nothing to send because messages
            # were filtered, or are aggregating values (e.g., mean policy).
            messagesToPost = []
            for message in messages:
                messagesFromPolicies = messagingPolicy.applyPolicies(message)
                if messagesFromPolicies is not None:
                    messagesToPost.extend(messagesFromPolicies)
            if len(messagesToPost) > 0:
                self.send(messagesToPost)
        else:
            raise StateException(
                "DirectlyConnectedDeviceImpl.offer(): not activated")

    # TODO: Shouldn't this be a config default value for timeout ?
    # TODO: Comment from JCL "Receive will ignore -1 timeout when not long
    # TODO: polling"
    # TODO: -1 means to "USE THE DEFAULT TIMEOUTS"
    # TODO: When 0 is passed results in "infinite" timeout
    def receive(self, timeout = -1):
        """
        Get messages from the IOT Cloud Service
        :param timeout: the time to wait for a response from the cloud service
        :returns:
        :raises: :class:`.StateException` if device is not activated.
        """
        if not self.tam.isActivated:
            raise StateException("DirectlyConnectedDevice.receive(): not activated")

        # Convert timeout
        if timeout == 0:
            connectionTimeout = type(self).USE_DEFAULT_TIMEOUT_VALUE
        elif timeout < 0:
            connectionTimeout = 0
        else:
            connectionTimeout = timeout

        t0 = TimeManager.currentTimeMillis()
        requestMessage = self.sendReceiveImpl.receive(connectionTimeout)

        if requestMessage is None:
            return None

        path = requestMessage.getURL()
        if DevicePolicyManager.PATH != path:
            return requestMessage

        try:
            responseMessage = \
                DevicePolicyManager.getDevicePolicyManager(self)\
                    .policyChanged(self, requestMessage)

            self.send([responseMessage])
        except Exception as e:
            self.getLogger().warning(f'receive(): {e}')
        
        remainingTime = connectionTimeout - \
                        (TimeManager.currentTimeMillis() - t0)
        if remainingTime > 0:
            return self.receive(remainingTime)
        else:
            return requestMessage

    def close(self):
        """
        Close the `MessageDispatcher`
        :return:
        """
        DirectlyConnectedDeviceImpl.closeStorageConnection(self.secureConn)
        self.secureConn.close()


    # TODO: Implement MQTT
    def createSendReceiveImpl(self):

        # if self.__isMqtt:
        #     impl = MqttSendReceiveImpl(secureConnection)
        #     impl.initialize()
        #     return impl
        # else:
        return HttpSendReceiveImpl(self.secureConn)

    def __eq__(self, other):

        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, DirectlyConnectedDeviceImpl):
            return False

        if self.tam.getServerHost() != other.tam.getServerHost():
            return False

        return self.tam.getClientId() == other.tam.getClientId()

    def __hash__(self):
        return hash((self.tam.getServerHost(), self.tam.getClientId()))

    def getLogger(self):
        return logging.getLogger("iotcs.device.impl.DirectlyConnectedDeviceImpl")

    def dumpMessage(self, message):
        dataitems = None
        if isinstance(message, (DataMessage, AlertMessage,)):
            dataitems = message.getDataItems()
        else:
            dataitems = list()

        messagedump = ""
        comma = ""
        for dataitem in dataitems:
            value = dataitem.getValue()
            key = dataitem.getKey()
            messagedump = f'{messagedump}{comma}"{key}"={value}'
            comma = ","

        return messagedump


class GatewayDeviceImpl(DirectlyConnectedDeviceImpl):

    __slots__ = (
    )

    def __init__(self, path, passphrase):
        """

        :param path:
        :param passphrase:
        :raises: GeneralSecurityException
        """
        super().__init__(path, passphrase, isGatewayDevice=True)

    def registerDevice(self, hardwareId, metaData:dict, deviceModels, restricted=False):
        """
        Register an indirectly-connected device with the cloud service and specify whether
        the gateway device is required to have the appropriate credentials for activating
        the indirectly-connected device.

            The `restricted` parameter controls whether or not the client
            library is `required` to supply credentials for activating
            the indirectly-connected device. The client library will
            `always` supply credentials for an indirectly-connected
            device whose trusted assets have been provisioned to the client.
            If, however, the trusted assets of the indirectly-connected device
            have not been provisioned to the client, the client library can
            create credentials that attempt to restrict the indirectly connected
            device to this gateway device.

            Pass `True` for the `restricted` parameter
            to ensure the indirectly-connected device cannot be activated
            by this gateway device without presenting credentials. If `restricted`
            is `True`, the client library will provide credentials to the server.
            The server will reject the activation request if the indirectly connected
            device is not allowed to roam to this gateway device.

            Pass `False` to allow the indirectly-connected device to be activated
            without presenting credentials if the trusted assets of the
            indirectly-connected device have not been provisioned to the client.
            If `restricted` is `False`, the client library will provide
            credentials if, and only if, the credentials have been provisioned to the
            client. The server will reject the activation if credentials are required
            but not supplied, or if the provisioned credentials do not allow the
            indirectly connected device to roam to this gateway device.

            The `hardwareId` is a unique identifier within the cloud service
            instance and may not be `None`. If one is not present for the device,
            it should be generated based on other metadata such as: model, manufacturer,
            serial number, etc.

            The `metaData` Dictionary should typically contain all the standard
            metadata (the constants documented in this class) along with any other
            vendor defined metadata.

        :param restricted: indicate whether or not credentials are required for
                          activating the indirectly connected device
        :param hardwareId: an identifier unique within the Cloud Service instance
        :param metaData: The metadata of the device
        :param deviceModels: should contain the device model type URNs supported by the indirectly connected device.
                       The {:code deviceModels} parameter is one or more, comma separated, device model URNs.
        :return: The endpoint id of the indirectly-connected device
        :raise: IOError if the message could not be sent
        :raise: ValueError  when key or signature algorithm class
                cannot be loaded, or the key is not in
                the trusted assets store, or the private key is invalid or
                if `hardwareId` is `None`,
                or `deviceModels` is `None` or zero length

        """
        if hardwareId is None or len(hardwareId) == 0:
            raise ValueError("hardwareId cannot be None or empty")

        if deviceModels is None or len(deviceModels) == 0:
            raise ValueError("at least one device model must be specified")

        # TODO: Why does JCL use getEndpointId() ?
        #data = hardwareId.encode("utf-8")
        data = self.getEndpointId().encode("utf-8")

        # If the ICD has been provisioned, use the shared secret to generate the
        # signature for the indirect activation request.
        signature = None
        try:
            # If this call throws a TrustException, then the ICD has not been provisioned.
            signature = self.tam.signWithSharedSecret(data, hardwareId)
        except Exception as e:
            # TODO: Need to discriminate as to the exception.
            # TODO: JCL is "TrustException" and e.getCause == null, re-throw
            # Assume it is the "ICD" not provisioned yet
            signature = None

        # If signature is None, then the ICD is not provisioned. But if the
        # restricted flag is True, then generate a signature which will cause the
        # ICD to be "locked" (for roaming) to the gateway
        if restricted and signature is None:
            signature = self.tam.signWithPrivateKey(data)


        indirectActivationRequest = createIndirectActivationRequest(hardwareId,
                                        metaData,
                                        deviceModels,
                                        signature)

        response = postIndirectActivationRequest(
            self.secureConn, indirectActivationRequest)

        return response.getEndpointId()




