# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import datetime
import json
import logging
from pathlib import Path
import sqlite3
import threading

from iotcs.common import TimeManager
from iotcs.common import Util
from iotcs.exception import StateException
from iotcs.message import AlertMessage
from iotcs.message import DataMessage
from iotcs.message import Message
from iotcs.message import ResponseMessage
from iotcs.message import RequestMessage
from iotcs.message import ResourceMessage

# com.oracle.iot.client.device.persistence

_UUID_LENGTH = 256
_ENDPOINT_ID_LENGTH = 256
_MESSAGE_ID_LENGTH = 256

def _getLogger():
    return logging.getLogger("iotcs.messaging.client.device.persistence")

def _getDbPath(dbname):
    """

    :param section:
    :param property:
    :param defaultDbname:
    :return:
    """
    dbpath = Path(dbname)
    if dbpath.is_absolute():
        return dbpath
    else:
        home = Path.home()
        return home.joinpath(str(dbname))

def _mkDbDir(dbpath):
    """

    :param dbpath: a Path to the db file
    :return:
    """
    # make all the directories leading up to the db file
    parentdirs = dbpath.parents[0]
    parentdirs.mkdir(parents=True, exist_ok=True)

def _deletedb(dbpath):
    try:
        if dbpath.exists():
            dbpath.unlink()
    except Exception as e:
        _getLogger().warning(f'_deletedb(): Could not delete {dbpath}, {e}')


def _getMessage(jsonMsg):
    """
    :param jsonMsg: a message in json format as a bytearray
    :raises: :class:`.StateException` if required properties are not present
    in `jsonMsg`
    """

    jsonObject = json.loads(jsonMsg)
    msgTypeStr = None
    try:
        msgTypeStr = jsonObject["type"]
    except Exception as e:
        raise StateException(f'Persistence.getMessage(): '
                             f'Message.Type "type" property is required.')

    msgType = None
    try:
        msgType = Message.Type[msgTypeStr]
    except Exception as e:
        raise StateException(f'Persistence.getMessage(): '
                             f'Invalid Message.Type {msgTypeStr}.')

    message = None
    try:
        if msgType is Message.Type.DATA:
            message = DataMessage().fromJson(jsonObject)
        elif msgType is Message.Type.REQUEST:
            message = RequestMessage().fromJson(jsonObject)
        elif msgType is Message.Type.RESPONSE:
            message = ResponseMessage().fromJson(jsonObject)
        elif msgType is Message.Type.RESOURCE:
            message = ResourceMessage().fromJson(jsonObject)
        elif msgType is Message.Type.ALERT:
            message = AlertMessage().fromJson(jsonObject)
    except Exception as e:
        raise e

    if message is None:
        raise StateException(f'Persistence.getMessage(): '
                             f'Invalid Message.Type {msgTypeStr}.')
    return message

class BatchByPersistence():
    """
    A class for persisting *DeviceFunction.BatchBy* policy data.

    See :class:`.BatchBySize`
    See :class:`.BatchByTime`
    See :class:`.BatchByCost`
    """

    _INSTANCE = None
    _PERSISTENCE_ENABLED = Util.getConfig()["batchby_persistence"] \
        .getboolean("enabled", fallback=False)

    TABLE_NAME = "BATCH_BY"

    CREATE_TABLE_SQL = f'CREATE TABLE IF NOT EXISTS {TABLE_NAME} ' \
                       f'(TIMESTAMP BIGINT NOT NULL,' \
                       f'ENDPOINT_ID VARCHAR({_ENDPOINT_ID_LENGTH}) NOT NULL,' \
                       f'MESSAGE_ID VARCHAR({_MESSAGE_ID_LENGTH}) NOT NULL,' \
                       f'MESSAGE BLOB, PRIMARY KEY (MESSAGE_ID))'

    ENDPOINT_ID_INDEX_SQL = \
        f'CREATE INDEX IF NOT EXISTS endpoint_id ON {TABLE_NAME} (ENDPOINT_ID)'

    # Explicitly list the desired select columns to guarantee the order
    # sql query parameter the endpoint id
    LOAD_TABLE_SQL = \
        f'SELECT TIMESTAMP,ENDPOINT_ID,MESSAGE_ID,MESSAGE FROM {TABLE_NAME} ' \
        f'WHERE ENDPOINT_ID = ? ORDER BY TIMESTAMP'

    # sql query parameter endpoint id
    LOAD_MESSAGE_SQL = \
        f'SELECT MESSAGE FROM {TABLE_NAME} ' \
        f'WHERE ENDPOINT_ID = ? ORDER BY TIMESTAMP'

    # sql query parameters timestamp, uuid, endpointid, message
    SAVE_SQL = \
        f'INSERT INTO {TABLE_NAME} (TIMESTAMP,ENDPOINT_ID,MESSAGE_ID,MESSAGE) VALUES (?, ?, ?, ?)'

    # TODO: Use IN
    # sql query parameter UUID
    DELETE_SQL = f'DELETE FROM {TABLE_NAME} WHERE MESSAGE_ID = ?'

    DBREADLOCK = None
    DBWRITELOCK = None
    DBNAME = None

    def __init__(self):
        """
        Return an instance of BatchByPersistence.

        :raises: :class:`.StateException` if persistence is not enabled.
        """
        if not BatchByPersistence.isEnabled():
            raise StateException(f'BatchByPersistence.__init__(): Persistence is not enabled.')

        dbname = Util.getConfig()["batchby_persistence"].get("db_name", fallback="iotcs/bbps.db")
        dbpath = _getDbPath(dbname)
        self.DBNAME = str(dbpath)
        _mkDbDir(dbpath)

        self.DBREADLOCK = threading.RLock()
        self.DBWRITELOCK = threading.RLock()

        with self.DBWRITELOCK:
            db = None
            try:
                db = sqlite3.connect(self.DBNAME)
                schema = f'{self.CREATE_TABLE_SQL};{self.ENDPOINT_ID_INDEX_SQL};'
                db.executescript(schema)
                db.commit()
            except Exception as e:
                _getLogger().critical(f'BatchByPersistence.__init__(): caught exception {e}')
            finally:
                if db is not None:
                    db.close()

    @classmethod
    def getInstance(cls):
        if cls.isEnabled():
            if cls._INSTANCE is None:
                cls._INSTANCE = BatchByPersistence()
            return cls._INSTANCE
        else:
            return None

    @classmethod
    def isEnabled(cls):
        return cls._PERSISTENCE_ENABLED


    @classmethod
    def __reset(cls):
        cls._INSTANCE = None
        dbpath = _getDbPath(Util.getConfig()["batchby_persistence"]
                            .get("db_name", fallback="iotcs/bbps.db"))
        _deletedb(dbpath)


    def save(self, messages, endpointId):
        """
        Save the given collection of messages to persistent storage.

        `None` or empty collections are supported, just ignored. `None` entries in
        the collection are ignored. This method will block until the job is
        complete.

        :param messages: The messages to save
        :param endpointId: Id of the subsystem that the messages belongs to
        """
        with self.DBWRITELOCK:
            self._save(messages, endpointId)

    def _save(self, messages, endpointId):
        db = None

        try:
            db = sqlite3.connect(self.DBNAME)
            sqlsave = self.SAVE_SQL.format(self.TABLE_NAME)
            # In case an exception happens before processing messages
            timestamp = ""
            messageid = ""
            messageJson = ""
            for message in messages:
                timestamp = message.getEventTime()
                if timestamp is None:
                    timestamp = TimeManager.currentTimeMillis(datetime.datetime.now())
                messageid = message.getClientId()

                # Not a string, a dict
                messageJson = json.dumps(message.toJson())
                # get bytearray
                msgblob = messageJson.encode()

                db.execute(sqlsave, (timestamp, endpointId, messageid, msgblob))

            db.commit()
        except Exception as e:
            if db is not None:
                db.rollback()
            _getLogger().warning(f'BatchByPersistence.save(): cannot update {self.TABLE_NAME} '
                                 f'with message clientId={messageid}, endpointId={endpointId}, '
                                 f'eventTime={TimeManager.getFormattedTime(timestamp)}, '
                                 f'message={messageJson}, exception={e}')
        finally:
            if db is not None:
                db.close()

    def delete(self, messages):
        """
        Delete the given collection of messages from persistent storage.

        `None` empty collections are supported, just ignored. `None` entries in the
        collection are ignored. This method will
        block until the job is complete.

        :param message: The messages to delete
        """
        # Just return if None or empty
        if messages is None or len(messages) == 0:
            return

        with self.DBWRITELOCK:
            self._delete(messages)


    def _delete(self, messages):
        db = None
        try:
            # TODO: Should this open and close for each delete ?
            # TODO: construct one large script using "IN"
            db = sqlite3.connect(self.DBNAME)
            for message in messages:
                # TODO: Should this really be tested for this condition ?
                # TODO: Need to find out what happens if such a row doesn't exist ?
                # TODO: Won't need in an "IN" clause.
                if message.getReliability() is Message.Reliability.GUARANTEED_DELIVERY:
                    continue

                messageid = message.getClientId()
                # Should only return one row
                db.execute(self.DELETE_SQL, (messageid,))

            db.commit()
        except Exception as e:
            _getLogger().warning(f'BatchByPersistence.delete(): cannot delete {self.TABLE_NAME} message'
                                 f'message={message}, exception={e}')
        finally:
            if db is not None:
                db.close()

    def load(self, endpointId):
        """
        Return a `list` of persisted messages associated with `endpointId`.

        :param endpointId: Id of the subsystem that the caller want to load
            messages for
        :return: a `list` messages that were persisted for given subsystem.
            Should not return `None` (it would be an error to do so)
            but `None` returns are treated as empty.
        """
        with self.DBREADLOCK:
            return self._load(endpointId)

    def _load(self, endpointId):
        db = None
        try:
            done = False
            db = sqlite3.connect(self.DBNAME)
            cur = db.cursor()
            # Should only return one row
            cur.execute(self.LOAD_MESSAGE_SQL, (endpointId,))
            messages = list()
            while not done:
                row = cur.fetchone()
                if row is None:
                    done = True
                    break
                # We only need MESSAGE
                msgblob = row[0]
                message = _getMessage(msgblob)
                if message is None:
                    _getLogger().warning(f'BatchByPersistence.load(): '
                                         f'cannot create message from {msgblob.decode()}')
                else:
                    messages.append(message)

            return messages
        except Exception as e:
            _getLogger().warning(f'BatchByPersistence.load(): '
                                 f'cannot load message for {endpointId}, {e}')
        finally:
            if db is not None:
                db.close()

class MessagePersistence():
    """A class for persisting :mod:`.message` messages."""

    _INSTANCE = None
    _PERSISTENCE_ENABLED = Util.getConfig()["message_persistence"]\
        .getboolean("enabled", fallback=False)

    TABLE_NAME = "MESSAGE_PERSISTENT_STORE"

    # TODO: Should be in external file
    # <tablename>, <maxuuidlength>, <maxendpointidlength>
    CREATE_TABLE_SQL = f'CREATE TABLE IF NOT EXISTS {TABLE_NAME} ' \
                       f'(TIMESTAMP BIGINT NOT NULL,' \
                       f'UUID VARCHAR({_UUID_LENGTH}) NOT NULL,' \
                       f'ENDPOINT_ID VARCHAR({_ENDPOINT_ID_LENGTH}) NOT NULL,' \
                       f'MESSAGE BLOB, PRIMARY KEY (UUID))'

    # <tablename>
    ENDPOINT_ID_INDEX_SQL = \
        f'CREATE INDEX IF NOT EXISTS endpoint_id ON {TABLE_NAME} (ENDPOINT_ID)'

    # Explicitly list the desired select columns to guarantee the order
    # <tablename>
    # sql query parameter the endpoint id
    LOAD_TABLE_SQL = \
        f'SELECT TIMESTAMP,UUID,ENDPOINT_ID,MESSAGE FROM {TABLE_NAME} ' \
        f'WHERE ENDPOINT_ID = ? ORDER BY TIMESTAMP'

    # <tablename>
    # sql query parameter endpoint id
    LOAD_MESSAGE_SQL = \
        f'SELECT MESSAGE FROM {TABLE_NAME} ' \
        f'WHERE ENDPOINT_ID = ? ORDER BY TIMESTAMP'

    # <tablename>
    # sql query parameters timestamp, uuid, endpointid, message
    SAVE_SQL = \
        f'INSERT INTO {TABLE_NAME} (TIMESTAMP,UUID,ENDPOINT_ID,MESSAGE) VALUES (?, ?, ?, ?)'

    # TODO: Make this delete like BatchByPersistence with where clause using
    # "IN" a set of UUID's to delete.
    # <tablename>
    # sql query parameter UUID
    DELETE_SQL = f'DELETE FROM {TABLE_NAME} WHERE UUID = ?'

    DBREADLOCK = None
    DBWRITELOCK = None
    DBNAME = None

    def __init__(self):
        """
        Return an instance of MessagePersistence.

        :raises: :class:`.StateException` if persistence is not enabled.
        """
        if not MessagePersistence.isEnabled():
            raise StateException(f'MessagesPersistence.__init__(): Persistence is not enabled.')

        dbname = Util.getConfig()["message_persistence"].get("db_name", fallback="iotcs/msps.db")
        dbpath = _getDbPath(dbname)
        self.DBNAME = str(dbpath)
        _mkDbDir(dbpath)

        self.DBREADLOCK = threading.RLock()
        self.DBWRITELOCK = threading.RLock()

        with self.DBWRITELOCK:
            db = None
            try:
                db = sqlite3.connect(self.DBNAME)
                schema = f'{self.CREATE_TABLE_SQL};{self.ENDPOINT_ID_INDEX_SQL};'
                db.executescript(schema)
                db.commit()
            except Exception as e:
                _getLogger().critical(f'MessagePersistence.__init__(): caught exception {e}')
            finally:
                if db is not None:
                    db.close()

    @classmethod
    def getInstance(cls):
        if cls.isEnabled():
            if cls._INSTANCE is None:
                cls._INSTANCE = MessagePersistence()
            return cls._INSTANCE
        else:
            return None

    @classmethod
    def isEnabled(cls):
        return cls._PERSISTENCE_ENABLED


    @classmethod
    def __reset(cls):
        cls._INSTANCE = None
        dbpath = _getDbPath(Util.getConfig()["message_persistence"]
                            .get("db_name", fallback="iotcs/msps.db"))
        _deletedb(dbpath)

    def save(self, messages, endpointId):
        """
        Save the given collection of messages to persistent storage.

        `None` or empty collections are supported, just ignored. `None` entries in
        the collection are ignored. This method will block until the job is
        complete.

        :param message: The messages to save
        :param endpointId: Id of the subsystem that the messages belongs to
        """
        with self.DBWRITELOCK:
            self._save(messages, endpointId)

    def _save(self, messages, endpointId):
        db = None

        try:
            db = sqlite3.connect(self.DBNAME)
            sqlsave = self.SAVE_SQL.format(self.TABLE_NAME)
            # In case an exception happens before processing messages
            _uuid = ""
            timestamp = ""
            messageJson = ""
            for message in messages:
                timestamp = message.getEventTime()
                if timestamp is None:
                    timestamp = TimeManager.currentTimeMillis(datetime.datetime.now())
                _uuid = message.getClientId()

                # Not a string, a dict
                messageJson = json.dumps(message.toJson())
                # get bytearray
                msgblob = messageJson.encode()

                db.execute(sqlsave, (timestamp, _uuid, endpointId, msgblob))

            db.commit()
        except Exception as e:
            if db is not None:
                db.rollback()
            _getLogger().warning(f'MessagePersistence.save(): cannot update {self.TABLE_NAME} '
                                 f'with message clientId={_uuid}, endpointId={endpointId}, '
                                 f'eventTime={TimeManager.getFormattedTime(timestamp)}, '
                                 f'message={messageJson}, exception={e}')
        finally:
            if db is not None:
                db.close()

    def delete(self, messages):
        """
        Delete the given collection of messages from persistent storage.
        
        `None` empty collections are supported, just ignored. `None` entries in the
        collection are ignored. This method will
        block until the job is complete.

        :param message: The messages to delete
        """
        # Just return if None or empty
        if messages is None or len(messages) == 0:
            return

        with self.DBWRITELOCK:
            self._delete(messages)

    def _delete(self, messages):
        db = None
        try:
            # TODO: Should this open and close for each delete ?
            # TODO: construct one large script using "IN"
            db = sqlite3.connect(self.DBNAME)
            for message in messages:
                _uuid = message.getClientId()
                # Should only return one row
                db.execute(self.DELETE_SQL, (_uuid,))

            db.commit()
        except Exception as e:
            _getLogger().warning(f'MessagePersistence.delete(): cannot delete {self.TABLE_NAME} message'
                                 f'message={message}, exception={e}')
        finally:
            if db is not None:
                db.close()

    def load(self, endpointId):
        """
        Return a `list` of persisted messages associated with `endpointId`.
        
        :param endpointId: Id of the subsystem that the caller want to load
            messages for
        :return: a `list` messages that were persisted for given subsystem.
            Should not return `None` (it would be an error to do so)
            but `None` returns are treated as empty.
        """
        with self.DBREADLOCK:
            return self._load(endpointId)

    def _load(self, endpointId):
        db = None
        try:
            done = False
            db = sqlite3.connect(self.DBNAME)
            cur = db.cursor()
            # Should only return one row
            cur.execute(self.LOAD_MESSAGE_SQL, (endpointId,))
            messages = list()
            while not done:
                row = cur.fetchone()
                if row is None:
                    done = True
                    break
                # We only need MESSAGE
                #timestamp = row["TIMESTAMP"]
                #clientId = row["UUID"]
                #endpointId = row["ENDPOINT_ID"]
                msgblob = row[0]
                message = _getMessage(msgblob)
                if message is None:
                    _getLogger().warning(f'MessagePersistence.load(): '
                        f'cannot create message from {msgblob.decode()}')
                else:
                    messages.append(message)

            return messages
        except Exception as e:
            _getLogger().warning(f'MessagePersistence.load(): '
                f'cannot load message for {endpointId}, {e}')
        finally:
            if db is not None:
                db.close()

