# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC
from abc import abstractmethod
from enum import Enum
import logging
import math
import numbers
import re

from iotcs.exception import ArgumentException
from iotcs.common import Util

# com.oracle.iot.shared

class Formula():
    """A class to represent a formula expresion."""

    tree = None
    # A :class:`.FormulaParser.Node` root of the AST.

    # TODO: Do we really need to be able to create an instance of this class ?
    def __init__(self, formula):
        """
        Create a :class:`.Formula` instance for `formula`.

        :param formula: a formula expression
        """
        tokens = FormulaParser.tokenize(formula)
        tree = FormulaParser.parseFormula(tokens, formula)

    @classmethod
    def compute(cls, vp, node=None):
        """
        Compute a value for :class:`.ValueProvider` `vp` using the parsed formula.

        :param vp: a :class:`.ValueProvider` instance
        :param node:
        :return: the computed value.
        """
        if node is None:
            node = cls.tree

        if node == None:
            return float('nan')
        if isinstance(node, (FormulaParser.Terminal, )):
            terminal = node
            attr = node.getValue()
            if terminal.terminaltype == FormulaParser.Terminal.Type.CURRENT_ATTRIBUTE:
                try:
                    value = vp.getCurrentValue(attr)
                    if isinstance(value, (numbers.Number, )):
                        return float(value)
                    elif isinstance(value, (bool, )):
                        return float(1) if value else float(0)
                    elif isinstance(value, (str, )):
                        return value
                except Exception as e:
                    cls.getLogger().warning(str(e))
            elif terminal.terminaltype == FormulaParser.Terminal.Type.IN_PROCESS_ATTRIBUTE:
                try:
                    value = vp.getInProcessValue(attr)
                    if value is None:
                        value = vp.getCurrentValue(attr)
                    if value is not None:
                        if isinstance(value, (numbers.Number, )):
                            return float(value)
                        elif isinstance(value, (bool, )):
                            return float(1) if value else float(0)
                        elif isinstance(value, (str, )):
                            return value
                except Exception as e:
                    cls.getLogger().warning(str(e))
            elif terminal.terminaltype == FormulaParser.Terminal.Type.NUMBER or \
                    terminal.terminaltype == FormulaParser.Terminal.Type.BOOL:
                try:
                    value = float(attr)
                    return value
                except Exception as e:
                    cls.getLogger().warning(str(e))
            elif terminal.terminaltype == FormulaParser.Terminal.Type.STRING or \
                    terminal.terminaltype == FormulaParser.Terminal.Type.IDENT:
                return attr
            return float('nan')
        if node.getOperation() == FormulaParser.Node.Operation.TERNARY:
            cond = float(cls.compute(vp, node.getLeftHandSide()))
            if cond == 1.0:
                return cls.compute(vp, node.getRightHandSide().getLeftHandSide())
            else:
                return cls.compute(vp, node.getRightHandSide().getRightHandSide())
        elif node.getOperation() == FormulaParser.Node.Operation.GROUP:
            return cls.compute(vp, node.getLeftHandSide())
        operation = node.getOperation()
        lhsComp = cls.compute(vp, node.getLeftHandSide())
        rhsComp = cls.compute(vp, node.getRightHandSide())
        if operation == FormulaParser.Node.Operation.FUNCTION:
            fn = (node.getLeftHandSide()).getValue()
            if "LOWER".lower() == fn.lower():
                str_ = str(rhsComp)
                return str_.lower() if str_ is not None else ""
            elif "UPPER".lower() == fn.lower():
                str_ = str(rhsComp)
                return str_.upper() if str_ is not None else ""
            else:
                cls.getLogger().warning(f'unknown function "{fn}"')
        elif operation == FormulaParser.Node.Operation.LOWER:
            return (str(lhsComp)).lower()
        elif operation == FormulaParser.Node.Operation.UPPER:
            return (str(lhsComp)).upper()
        elif operation == FormulaParser.Node.Operation.EQ:
            if lhsComp.__class__ == rhsComp.__class__:
                if isinstance(lhsComp, float):
                    if math.isnan(lhsComp) and math.isnan(rhsComp):
                        return 1.0
                return 1.0 if lhsComp == rhsComp else 0.0
            else:
                return 0.0
        elif operation == FormulaParser.Node.Operation.NEQ:
            if lhsComp.__class__ == rhsComp.__class__:
                if isinstance(lhsComp, float):
                    if math.isnan(lhsComp) and math.isnan(rhsComp):
                        return 0.0
                return 0.0 if lhsComp == rhsComp else 1.0
            else:
                return 1.0
        elif operation == FormulaParser.Node.Operation.PLUS:
            if isinstance(lhsComp, (str, )) and isinstance(rhsComp, (str, )):
                lhsString = str(lhsComp)
                rhsString = str(rhsComp)
                return f'{lhsString}{rhsString}'
        elif operation == FormulaParser.Node.Operation.LIKE:
            if isinstance(lhsComp, (str, )) and isinstance(rhsComp, (str, )):
                lhsString = str(lhsComp)
                rhsString = str(rhsComp)
                return 1.0 if cls.sqlMatches(lhsString, rhsString) else 0.0
            else:
                return 0.0
        lhs = float(lhsComp)
        rhs = float(rhsComp)
        if operation == FormulaParser.Node.Operation.UNARY_MINUS:
            return -lhs
        elif operation == FormulaParser.Node.Operation.UNARY_PLUS:
            return +lhs
        elif operation == FormulaParser.Node.Operation.DIV:
            return lhs / rhs
        elif operation == FormulaParser.Node.Operation.MUL:
            return lhs * rhs
        elif operation == FormulaParser.Node.Operation.PLUS:
            return lhs + rhs
        elif operation == FormulaParser.Node.Operation.MINUS:
            return lhs - rhs
        elif operation == FormulaParser.Node.Operation.MOD:
            return lhs % rhs
        elif operation == FormulaParser.Node.Operation.OR:
            #  Let NaN or NaN be false
            if math.isnan(lhs):
                return 0.0 if math.isnan(rhs) else 1.0
            return 1.0 if lhs != 0.0 or rhs != 0.0 else 0.0
        elif operation == FormulaParser.Node.Operation.AND:
            #  If lhs or rhs is NaN, return false
            if math.isnan(lhs) or math.isnan(rhs):
                return 0.0
            return 1.0 if lhs != 0.0 and rhs != 0.0 else 0.0
        elif operation == FormulaParser.Node.Operation.GT:
            #  NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
            #  Let NaN > 42 return false, and 42 > NaN return true
            if math.isnan(lhs):
                return 0.0
            if math.isnan(rhs):
                return 1.0
            return 1.0 if lhs > rhs else 0.0
        elif operation == FormulaParser.Node.Operation.GTE:
            #  NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
            #  Let NaN >= 42 return false, and 42 >= NaN return true
            if math.isnan(lhs):
                return 1.0 if math.isnan(rhs) else 0.0
            if math.isnan(rhs):
                return 1.0
            return 1.0 if lhs >= rhs else 0.0
        elif operation == FormulaParser.Node.Operation.LT:
            #  NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
            #  Let NaN < 42 return false, and 42 < NaN return true
            if math.isnan(lhs):
                return 0.0
            if math.isnan(rhs):
                return 1.0
            return 1.0 if lhs < rhs else 0.0
        elif operation == FormulaParser.Node.Operation.LTE:
            #  NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
            #  Let NaN <= 42 return false, and 42 <= NaN return true
            if math.isnan(lhs):
                return 1.0 if math.isnan(rhs) else 0.0
            if math.isnan(rhs):
                return 1.0
            return 1.0 if lhs <= rhs else 0.0
        elif operation == FormulaParser.Node.Operation.TERNARY:
            pass
        elif operation == FormulaParser.Node.Operation.ALTERNATIVE:
            pass
        elif operation == FormulaParser.Node.Operation.NOT:
            return 0.0 if lhs == 1.0 else 1.0
        elif operation == FormulaParser.Node.Operation.FUNCTION:
            pass
        elif operation == FormulaParser.Node.Operation.GROUP:
            pass
        elif operation == FormulaParser.Node.Operation.TERMINAL:
            pass
        return float('nan')

    @classmethod
    def sqlMatches(cls, s, regEx):
        """
        Return `True` if `s` matches regular expression `regEx`

        It is assumed that a literal back slash in regEx is escaped
        appropriately according to python regular expression rules.
        :param s: string to match
        :param regEx: a regular expression using SQL wildcards *%* or *_*
        """
        sb = f''
        prev = 0
        i = 0
        len_ = len(regEx)
        while i < len_:
            next = regEx[i]
            if next == '_':
                if prev == '\\':
                    sb = sb[:i-1] + sb[i:]
                else:
                    #  Covert _ to .
                    next = '.'
            elif next == '%':
                if prev == '\\':
                    sb = sb[:i-1] + sb[i:]
                else:
                    #  Convert % to .*
                    sb = f'{sb}.'
                    next = '*'

            sb = f'{sb}{next}'
            prev = next
            i += 1


        sbre = re.compile(sb)
        return sbre.match(s) is not None

    @classmethod
    def getAttributeReferences(cls, formula):
        """
        Return the attribute referenced in `formula` *$(<attribute>)*.
        :param formula:
        :return: a `list` of attributes or `[]`
        """
        attributes = []
        if formula is None:
            return attributes
        # returns -1 if not found
        pos = formula.find("$(")
        while pos != -1:
            end = formula.find(')', pos + 1)
            # check for '$$'
            if pos == 0 or formula[pos - 1] != '$':
                attr = formula[pos + len('$('): end]
                attributes.append(attr)
            pos = formula.indexOf("$(", end + 1)

        return attributes

    def dump(self):
        """ generated source for method dump """
        return FormulaParser.dump(self.tree)

    @classmethod
    def getLogger(cls):
        """ generated source for method getLogger """
        return logging.Logger("iotcs.messaging.shared.Formula")

# com.oracle.iot.shared

class FormulaParser():
    """A Formula parser."""

    class Token():
        """The Lexer."""
        # TODO: enum ?
        class Type(Enum):
            """FormulaParser lexer token types."""
            AND = 0 #u'AND'
            #  &&
            COLON = 1 #u'COLON'
            #  :
            COMMA = 2 #u'COMMA'
            #  ,
            DIV = 3 #u'DIV'
            #  \
            DOLLAR = 4 #u'DOLLAR'
            #  $
            EQ = 5 #u'EQ'
            #  =
            FUNCTION = 6 #u'FUNCTION'
            #  IDENT '('
            ATTRIBUTE = 7 #u'ATTRIBUTE'
            #  '$(' IDENT ')'
            GT = 8 #u'GT'
            #  >
            GTE = 9 #u'GTE'
            #  >=
            IDENT = 10 #u'IDENT'
            #  [_a-zA-Z]+ [_a-zA-Z0-9\-]*
            LPAREN = 11 #u'LPAREN'
            #  (
            LT = 12 #u'LT'
            #  <
            LTE = 13 #u'LTE'
            #  <=
            MINUS = 14 #u'MINUS'
            #  -
            MOD = 15 #u'MOD'
            #  %
            MUL = 16 #u'MUL'
            #  *
            NEQ = 17 #u'NEQ'
            #  !=
            NOT = 18 #u'NOT'
            #  !
            NUMBER = 19 #u'NUMBER'
            #  [0-9]+|[0-9]*"."[0-9]+
            OR = 20 #u'OR'
            #  ||
            PLUS = 21 #u'PLUS'
            #  +
            QUESTION_MARK = 22 #u'QUESTION_MARK'
            #  ?
            RPAREN = 23 #u'RPAREN'
            #  )
            STRING = 24 #u'STRING'
            #  "
            LIKE = 25 #u'LIKE'
            #  LIKE String.matches()
            WS = 26 #u'WS'
            #  whitespace is not significant and is consumed
            BOOL = 27
            # ignorecase(TRUE) or ignorecase(FALSE)

        __slots__ = (
            'tokentype',
            'pos',
            'length'
        )

        def __init__ (self, tokentype, pos, length):
            self.tokentype = tokentype
            self.pos = pos
            self.length = length

        def getType(self):
            return self.tokentype


        def getPos(self):
            return self.pos

        def getLength(self):
            return self.length

        def __eq__(self, other):
            if self is other:
                return True
            if other is None or other.__class__ is not self.__class__:
                return False

            return self.tokentype == other.tokentype and \
                self.pos == other.pos and \
                self.length == other.length

    @classmethod
    def tokenize(cls, formula):
        """Tokenize the formula string."""
        tokens = []
        pos = 0
        tokenType = None
        inString = False
        i = 0
        while i < len(formula):
            lasttokentype = tokenType
            length = i - pos
            ch = formula[i]
            if inString and ch != '"':
                i += 1
                continue 
            if ch == '"':
                if inString:
                    inString = False
                    i += 1
                    continue 
                else:
                    lasttokentype = cls.Token.Type.STRING
                    inString = True
            elif ch == '(':
                lasttokentype = cls.Token.Type.LPAREN
            elif ch == ')':
                lasttokentype = cls.Token.Type.RPAREN
            elif ch == ',':
                lasttokentype = cls.Token.Type.COMMA
            elif ch == '?':
                lasttokentype = cls.Token.Type.QUESTION_MARK
            elif ch == ':':
                lasttokentype = cls.Token.Type.COLON
            elif ch == '+':
                lasttokentype = cls.Token.Type.PLUS
            elif ch == '-':
                if tokenType != cls.Token.Type.IDENT:
                    lasttokentype = cls.Token.Type.MINUS
            elif ch == '*':
                lasttokentype = cls.Token.Type.MUL
            elif ch == '/':
                lasttokentype = cls.Token.Type.DIV
            elif ch == '%':
                lasttokentype = cls.Token.Type.MOD
            elif ch == '=':
                lasttokentype = cls.Token.Type.EQ
                peekChar = cls.peekformula(formula, i + 1)
                #  be forgiving of '=='
                if peekChar == '=':
                    i += 1
            elif ch == '!':
                peekChar = cls.peekformula(formula, i + 1)
                if peekChar == '=':
                    lasttokentype = cls.Token.Type.NEQ
                    i += 1
                else:
                    lasttokentype = cls.Token.Type.NOT
            elif ch == '>':
                peekChar = cls.peekformula(formula, i + 1)
                if peekChar == '=':
                    lasttokentype = cls.Token.Type.GTE
                    i += 1
                else:
                    lasttokentype = cls.Token.Type.GT
            elif ch == '<':
                peekChar = cls.peekformula(formula, i + 1)
                if peekChar == '=':
                    lasttokentype = cls.Token.Type.LTE
                    i += 1
                else:
                    lasttokentype = cls.Token.Type.LT
            elif ch == '|':
                peekChar = cls.peekformula(formula, i + 1)
                if peekChar == '|':
                    lasttokentype = cls.Token.Type.OR
                    i += 1
            elif ch == '&':
                peekChar = cls.peekformula(formula, i + 1)
                if peekChar == '&':
                    lasttokentype = cls.Token.Type.AND
                    i += 1
            elif ch == '$':
                peekChar = cls.peekformula(formula, i + 1)
                if peekChar == '(':
                    lasttokentype = cls.Token.Type.ATTRIBUTE
                    i += 1
                else:
                    lasttokentype = cls.Token.Type.DOLLAR
            else:
                if ch.isspace():
                    lasttokentype = cls.Token.Type.WS
                elif tokenType != cls.Token.Type.IDENT:
                    if ch.isdigit():
                        lasttokentype = cls.Token.Type.NUMBER
                    elif ch == '.':
                        #  [0-9]+|[0-9]*"."[0-9]+
                        if tokenType != cls.Token.Type.NUMBER:
                            peekChar = cls.peekformula(formula, i + 1)
                            if peekChar.isigit():
                                lasttokentype = cls.Token.Type.NUMBER
                                i += 1
                            else:
                                raise ArgumentException(
                                    f'Found "{peekChar}" @ {i + 1} : expected [0-9]')
                    else:
                        lasttokentype = cls.Token.Type.IDENT
            #  Add previous token when lexer hits a new token.
            if tokenType != lasttokentype or \
                    (tokenType == cls.Token.Type.LPAREN and
                        lasttokentype == cls.Token.Type.LPAREN) or \
                    (tokenType == cls.Token.Type.RPAREN and
                        lasttokentype == cls.Token.Type.RPAREN):
                if tokenType == cls.Token.Type.IDENT:
                    token = formula[pos:pos + length]
                    if "AND".lower() == token.lower():
                        tokenType = cls.Token.Type.AND
                    elif "OR".lower() == token.lower():
                        tokenType = cls.Token.Type.OR
                    elif "NOT".lower() == token.lower():
                        tokenType = cls.Token.Type.NOT
                    elif "LIKE".lower() == token.lower():
                        tokenType = cls.Token.Type.LIKE
                    elif "TRUE".lower() == token.lower() or \
                            "FALSE".lower() == token.lower():
                        tokenType = cls.Token.Type.BOOL
                    elif lasttokentype == cls.Token.Type.LPAREN:
                        tokenType = lasttokentype = cls.Token.Type.FUNCTION
                        i += 1
                        continue 
                #  tokenType should only be null the first time through
                if tokenType is not None:
                    if tokenType != cls.Token.Type.WS:
                        tokens.append(cls.Token(tokenType, pos, length))
                    pos += length
                #  previous token is now current token
                tokenType = lasttokentype

            i += 1

        #  add the last token
        if tokenType != cls.Token.Type.WS:
            # Special case, TRUE or FALSE as the only element in the
            # expression
            if tokenType == cls.Token.Type.IDENT:
                token = formula[pos:len(formula) - pos]
                if "TRUE".lower() == token.lower() or \
                     "FALSE".lower() == token.lower():
                    tokenType = cls.Token.Type.BOOL

            tokens.append(cls.Token(tokenType, pos, len(formula) - pos))
        return tokens

    @classmethod
    def peekformula(cls, s, offset):
        """
        Return the next character in `s` from `offset`.

        :param s: return the charcter at `offset`
        :param offset: the character index in `s` to return
        """
        return s[offset] if (offset < len(s)) else None //'\0'


    class Node():
        """The FormulaParser AST."""
        class Operation:
            """Expression operators."""
            #  terminal is a number or attribute, LHS is a Terminal, no RHS
            __slots__ = (
                'prec',
                'op'
            )

            UNARY_PLUS = None
            UNARY_MINUS = None
            PLUS = None
            MINUS = None
            MUL = None
            DIV = None
            MOD = None
            AND = None
            OR = None
            EQ = None
            LIKE = None
            NEQ = None
            LT = None
            LTE = None
            GT = None
            GTE = None
            TERNARY = None
            ALTERNATIVE = None
            NOT = None
            LOWER = None
            UPPER = None
            FUNCTION = None
            GROUP = None
            TERMINAL = None

            def __init__(self, precedence, op):
                """
                Create an :class:`.Operation` instance.

                :param precedence: the operator precedence
                """
                self.prec = precedence
                self.op = op

            def getPrec(self):
                """Return the operator precedence."""
                return self.prec

            def __str__(self):
                return self.op


        Operation.UNARY_PLUS = Operation(6, "UNARY_PLUS")
        Operation.UNARY_MINUS = Operation(6, "UNARY_MINUS")
        Operation.PLUS = Operation(3, "PLUS")
        Operation.MINUS = Operation(3, "MINUS")
        Operation.MUL = Operation(4, "MUL")
        Operation.DIV = Operation(4, "DIV")
        Operation.MOD = Operation(4, "MOD")
        Operation.AND = Operation(1, "AND")
        Operation.OR = Operation(1, "OR")
        Operation.EQ = Operation(2, "EQ")
        Operation.LIKE = Operation(2, "LIKE")
        Operation.NEQ = Operation(2, "NE")
        Operation.LT = Operation(2, "LT")
        Operation.LTE = Operation(2, "LTE")
        Operation.GT = Operation(2, "GT")
        Operation.GTE = Operation(2, "GTE")
        Operation.TERNARY = Operation(0, "TERNARY")
        Operation.ALTERNATIVE = Operation(0, "ALTERNATIVE")
        Operation.NOT = Operation(6, "NOT")
        Operation.LOWER = Operation(6, "LOWER")
        Operation.UPPER = Operation(6, "UPPER")
        Operation.FUNCTION = Operation(6, "FUNCTION")
        Operation.GROUP = Operation(-1, "GROUP")
        Operation.TERMINAL = Operation(-1, "TERMINAL")

        __slots__ = (
            'operation',
            'leftHandSide',
            'rightHandSide'
        )

        # Note: JCL does not include a "rightHandSide" argument and uses
        # a java syntactic feature to set the right hand side, not available
        # in python, so define the argument, and default it to None.
        def __init__(self, operation, leftHandSide, rightHandSide=None):
            """
            Create a `Node` instance of `operation`.

            :param operation: the operator represented by this `Node`
            :param leftHandSide: the left hand side `Node`.
            """
            self.operation = operation
            self.leftHandSide = leftHandSide
            self.rightHandSide = rightHandSide

        def getOperation(self):
            """Return the :class:`.Operation` for this `Node`."""
            return self.operation

        def getLeftHandSide(self):
            """
            Return the left hand side `Node`.
            :return: the left hand side 'Node`
            """
            return self.leftHandSide

        def setLeftHandSide(self, leftHandSide):
            """
            Set the left hand side `Node`.

            :param leftHandSide: the left hand side `Node`.
            """
            self.leftHandSide = leftHandSide

        def getRightHandSide(self):
            """
            Return the right hand side `Node`.
            :return: the right hand side `Node`
            """
            return self.rightHandSide

        def setRightHandSide(self, rightHandSide):
            """Set the right hand side `Node`."""
            self.rightHandSide = rightHandSide

        def __eq__(self, other):
            """
            Return `True` if `self` equals `other`.

            :param other: an instance of `Node`
            :return: `True` if `self` equals `other` else `False`
            """
            if self is other:
                return True
            if other is None or other.__class__ != self.__class__:
                return False

            if self.leftHandSide is None:
                if other.leftHandSide is not None:
                    return False
            elif other.leftHandSide is None:
                return False

            if self.rightHandSide is None:
                if other.rightHandSide is not None:
                    return False
            elif other.rightHandSide is None:
                return False

            return self.leftHandSide == other.leftHandSide and \
                   self.rightHandSide == other.rightHandSide

    class Terminal(Node):
        """
        A terminal is either an attribute, number, or no-arg function.

        It is a Node, but has no Operation, LHS, or RHS
        """
        # TODO: Why not enum ?
        class Type(Enum):
            """Terminal Type"""
            IN_PROCESS_ATTRIBUTE = 0 #u'IN_PROCESS_ATTRIBUTE'
            CURRENT_ATTRIBUTE = 1 #u'CURRENT_ATTRIBUTE'
            NUMBER = 2 #u'NUMBER'
            IDENT = 3 #u'IDENT'
            STRING = 4 #u'STRING'
            BOOL = 5

        __slots__ = (
            'terminaltype',
            'value'
        )

        def __init__(self, terminaltype, value):
            """Create a Terminal instance."""
            super(FormulaParser.Terminal, self).__init__(
                FormulaParser.Node.Operation.TERMINAL, None)
            self.terminaltype = terminaltype
            self.value = value

        def getValue(self):
            """
            Return the `terminal value.
            :return: the terminal value.
            """
            return self.value

        def __eq__(self, other):
            """
            Return `True` if `self` equals `other`
            :param other:
            :return:
            """
            if self is other:
                return True
            if other is None or other.__class__ != self.__class__:
                return False

            if self.terminaltype != other.terminaltype:
                return False

            if self.value is not None:
                return self.value == other.value
            else:
                return other.value is None

    @classmethod
    def comparePrec(cls, left, right):
        """
        Return 0 if `left` and `right` are of equal precedence.

        Return a value less than zero if `left` has lower precedence than
        `right` else greater than zero if `left` has higher precedence than
        `left`
        :return: the precedence of `left` relative to `right`
        """
        return left.getOperation().getPrec() - right.getOperation().getPrec()


    @classmethod
    def parseFormula(cls, tokens, formula):
        """
        Parse the formula into tokens.

        Formula::

            formula
                : numericExpression
                | ternaryExpression
                ;


        :param tokens: the tokens of formula
        :param formula: the formula to parse.

        :return: the root of the AST
        """
        stack = []
        index = -1
        try:
            index = cls.parseNumericExpression(stack, tokens, formula, 0)
        except ArgumentException as e:
            pass
            #  drop through = try as conditional expression
        if index < len(tokens):
            stack.clear()
            try:
                index = cls.parseTernaryExpression(stack, tokens, formula, 0)
            except ArgumentException as e:
                pass
                #  drop through = try as conditional expression
            except IndexError as e:
                pass
                #  drop through = try as conditional expression
        if index < len(tokens):
            stack.clear()
            index = cls.parseRelationalExpression(stack, tokens, formula, 0)
        if index < len(tokens):
            lastToken = tokens[index]
            raise ArgumentException(f'formula: parser bailed @ {lastToken.pos}')
        return stack[0]

    @classmethod
    def parseTernaryExpression(cls, stack, tokens, formula, index):
        """
        Parse a ternary expression.

        TernaryExpression::

            ternaryExpression
                : conditionalOrExpression QUESTION_MARK additiveExpression COLON additiveExpression
                ;
        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        index = cls.parseConditionalOrExpression(stack, tokens, formula, index)
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        if token.getType() != cls.Token.Type.QUESTION_MARK:
            raise ArgumentException(
                f'parseTernaryExpression: found {str(token)}, '
                f'expected QUESTION_MARK')
        ternary = cls.Node(cls.Node.Operation.TERNARY, stack.pop())
        index = cls.parseAdditiveExpression(stack, tokens, formula, index + 1)
        token = tokens[index]
        if token.getType() != cls.Token.Type.COLON:
            raise ArgumentException(
                f'parseTernaryExpression: found {str(token)}, expected COLON')
        alternatives = cls.Node(cls.Node.Operation.ALTERNATIVE, stack.pop())
        ternary.setRightHandSide(alternatives)
        index = cls.parseAdditiveExpression(stack, tokens, formula, index + 1)
        alternatives.setRightHandSide(stack.pop())
        stack.append(ternary)
        return index


    # 
    #  returns the index of the next token to be processed.
    @classmethod
    def parseConditionalOrExpression(cls, stack, tokens, formula, index):
        """
        Parse a conditional expression.

        ConditionalExpression::
            conditionalOrExpression
                : conditionalAndExpression ( OR conditionalAndExpression )*
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        index = cls.parseConditionalAndExpression(stack, tokens, formula, index)
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        lhs = None
        if token.getType() == FormulaParser.Token.Type.OR:
            lhs = cls.Node(cls.Node.Operation.OR, stack.pop())
            index += 1
        else:
            return index
        index = cls.parseConditionalOrExpression(stack, tokens, formula, index)
        stack.append(cls.prioritize(lhs, stack.pop()))
        return index

    # 
    #  conditionalAndExpression
    #      : valueLogical ( AND valueLogical )*
    #      ;
    # 
    #  returns the index of the next token to be processed.
    @classmethod
    def parseConditionalAndExpression(cls, stack, tokens, formula, index):
        """
        Parse a conditional AND expression.

        ConditionalAndExpression::

            conditionalAndExpression
                : valueLogical ( AND valueLogical )*
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        index = cls.parseValueLogical(stack, tokens, formula, index)
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        lhs = None
        if token.getType() == FormulaParser.Token.Type.AND:
            lhs = cls.Node(cls.Node.Operation.AND, stack.pop())
            index += 1
        else:
            return index
        index = cls.parseConditionalAndExpression(stack, tokens, formula, index)
        stack.append(cls.prioritize(lhs, stack.pop()))
        return index

    @classmethod
    def parseValueLogical(cls, stack, tokens, formula, index):
        """
        Parse a logical value.

        ValueLogical::

            valueLogical
                : relationalExpression
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        return cls.parseRelationalExpression(stack, tokens, formula, index)

    # 

    # 
    #  returns the index of the next token to be processed.
    @classmethod
    def parseRelationalExpression(cls, stack, tokens, formula, index):
        """
        Parse a relational expression.

        RelationalExpression::

            relationalExpression
                : numericExpression (EQ numericExpression | NEQ numericExpression | LT numericExpression | GT numericExpression | LTE numericExpression | GTE numericExpression )?
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        index = cls.parseNumericExpression(stack, tokens, formula, index)
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        lhs = None
        if token.getType() == FormulaParser.Token.Type.LIKE:
            lhs = cls.Node(cls.Node.Operation.LIKE, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.EQ:
            lhs = cls.Node(cls.Node.Operation.EQ, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.NEQ:
            lhs = cls.Node(cls.Node.Operation.NEQ, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.LT:
            lhs = cls.Node(cls.Node.Operation.LT, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.LTE:
            lhs = cls.Node(cls.Node.Operation.LTE, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.GT:
            lhs = cls.Node(cls.Node.Operation.GT, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.GTE:
            lhs = cls.Node(cls.Node.Operation.GTE, stack.pop())
            index += 1
        else:
            return index
        index = cls.parseRelationalExpression(stack, tokens, formula, index)
        stack.append(cls.prioritize(lhs, stack.pop()))
        return index

    @classmethod
    def parseNumericExpression(cls, stack, tokens, formula, index):
        """
        Parse a numerical expression.

        NumericalExpression::
            numericExpression
                : additiveExpression
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        return cls.parseAdditiveExpression(stack, tokens, formula, index)

    @classmethod
    def parseAdditiveExpression(cls, stack, tokens, formula, index):
        """
        Parse an additive expression.

        AdditiveExpression::
            additiveExpression
                :multiplicativeExpression (PLUS multiplicativeExpression | MINUS multiplicativeExpression )*
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        index = cls.parseMultiplicativeExpression(stack, tokens, formula, index)
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        lhs = None
        if token.getType() == FormulaParser.Token.Type.PLUS:
            lhs = cls.Node(cls.Node.Operation.PLUS, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.MINUS:
            lhs = cls.Node(cls.Node.Operation.MINUS, stack.pop())
            index += 1
        else:
            return index
        index = cls.parseAdditiveExpression(stack, tokens, formula, index)
        stack.append(cls.prioritize(lhs, stack.pop()))
        return index

    @classmethod
    def parseMultiplicativeExpression(cls, stack, tokens, formula, index):
        """
        Parse a multiplicative expression.

        MultiplicativeExpression::
            multiplicativeExpression
                : exponentiationExpression (MUL exponentiationExpression | DIV exponentiationExpression | MOD exponentiationExpression)*
                ;
        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        index = cls.parseUnaryExpression(stack, tokens, formula, index)
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        lhs = None
        if token.getType() == FormulaParser.Token.Type.MUL:
            lhs = cls.Node(cls.Node.Operation.MUL, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.DIV:
            lhs = cls.Node(cls.Node.Operation.DIV, stack.pop())
            index += 1
        elif token.getType() == FormulaParser.Token.Type.MOD:
            lhs = cls.Node(cls.Node.Operation.MOD, stack.pop())
            index += 1
        else:
            return index
        index = cls.parseMultiplicativeExpression(stack, tokens, formula, index)
        stack.append(cls.prioritize(lhs, stack.pop()))
        return index

    @classmethod
    def parseUnaryExpression(cls, stack, tokens, formula, index):
        """
        Parse a unary expression.

        UnaryExpression::
            unaryExpression
                : NOT primaryExpression
                | PLUS primaryExpression
                | MINUS primaryExpression
                | LOWER primaryExpression
                | UPPER primaryExpression
                | primaryExpression
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        if token.getType() == FormulaParser.Token.Type.NOT:
            index = cls.parsePrimaryExpression(stack, tokens, formula, index + 1)
            stack.append(cls.Node(cls.Node.Operation.NOT, stack.pop()))
        elif token.getType() == FormulaParser.Token.Type.PLUS:
            index = cls.parsePrimaryExpression(stack, tokens, formula, index + 1)
            stack.append(cls.Node(cls.Node.Operation.UNARY_PLUS, stack.pop()))
        elif token.getType() == FormulaParser.Token.Type.MINUS:
            index = cls.parsePrimaryExpression(stack, tokens, formula, index + 1)
            stack.append(cls.Node(cls.Node.Operation.UNARY_MINUS, stack.pop()))
        else:
            index = cls.parsePrimaryExpression(stack, tokens, formula, index)
        return index

    @classmethod
    def parsePrimaryExpression(cls, stack, tokens, formula, index):
        """
        Parse a primary expression.

        PrimaryExpression::
            primaryExpression
                : brackettedExpression
                | functionElement
                | expressionElement
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        newIndex = cls.parseBrackettedExpression(stack, tokens, formula, index)
        if newIndex >= len(tokens):
            return len(tokens)
        if newIndex == index:
            newIndex = cls.parseFunctionElement(stack, tokens, formula, index)
            if newIndex == index:
                newIndex = cls.parseExpressionElement(stack, tokens, formula, index)
        if newIndex == index:
            raise ArgumentException(
                f'parsePrimaryExpression: expected '
                f'[brackettedExpression|functionElement|expressionElement]'
        )
        return newIndex

    @classmethod
    def parseBrackettedExpression(cls, stack, tokens, formula, index):
        """
        Parse a bracketted expression.

        BrackettedExpression::
            brackettedExpression
                : LPAREN conditionalOrExpression RPAREN
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        if token.getType() == FormulaParser.Token.Type.LPAREN:
            index = cls.parseConditionalOrExpression(stack, tokens, formula,
                                                     index + 1)
            current = cls.peek(tokens, index)
            if current.getType() != cls.Token.Type.RPAREN:
                raise ArgumentException(
                    f'term: Found {current.getType()} @ '
                    f'{current.getPos()} expected RPAREN')
            stack.append(cls.Node(cls.Node.Operation.GROUP, stack.pop()))
            index += 1
            #  consume RPAREN
        return index

    @classmethod
    def parseFunctionElement(cls, stack, tokens, formula, index):
        """
        Parse a function element.

        FunctionElement::
            functionElement
                : FUNCTION (args)? RPAREN
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        if token.getType() == FormulaParser.Token.Type.FUNCTION:
            next = cls.peek(tokens, index + 1)
            #  token.getLength()-1 to strip off LPAREN
            value = formula[token.getPos():
                            token.getPos() + token.getLength() - 1]
            #  FUNCTION operation has function name on LHS, args
            # chaining from RHS to RHS
            function_ = cls.Node(cls.Node.Operation.FUNCTION,
                                 cls.Terminal(cls.Terminal.Type.IDENT, value))
            if next.getType() == cls.Token.Type.RPAREN:
                #  no-arg function
                pass
            else:
                #  FUNCTION arg [, arg]* )
                index = cls.parseArgs(stack, tokens, formula, index + 1)
                function_.setRightHandSide(stack.pop())
                current = cls.peek(tokens, index)
                if current.getType() != cls.Token.Type.RPAREN:
                    raise ArgumentException(
                        f'term: Found {current.getType()} @ '
                        f'{current.getPos()}. Expected RPAREN')
            stack.append(function_)
            index += 1
            #  consume RPAREN
        return index

    @classmethod
    def parseExpressionElement(cls, stack, tokens, formula, index):
        """
        Parse an expression element.

        ExpressionElement::
            expressionElement
                : STRING | IDENT | NUMBER | BOOL | propertyRef
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        if token.getType() == FormulaParser.Token.Type.STRING:
            #  Do not include quotes in the value
            value = formula[token.getPos() + 1:
                            token.getPos() + token.getLength() - 1]
            stack.append(cls.Terminal(cls.Terminal.Type.STRING, value))
            index += 1
            #  consume STRING
        elif token.getType() == FormulaParser.Token.Type.IDENT:
            value = formula[token.getPos():
                            token.getPos() + token.getLength()]
            stack.append(cls.Terminal(cls.Terminal.Type.IDENT, value))
            index += 1
            #  consume IDENT
        elif token.getType() == FormulaParser.Token.Type.NUMBER:
            value = formula[token.getPos():
                            token.getPos() + token.getLength()]
            stack.append(cls.Terminal(cls.Terminal.Type.NUMBER, value))
            index += 1
            #  consume NUMBER
        elif token.getType() == FormulaParser.Token.Type.BOOL:
            value = formula[token.getPos():
                            token.getPos() + token.getLength()]
            stack.append(cls.Terminal(cls.Terminal.Type.BOOL, Util.strtobool(value)))
            index += 1
            # consume BOOL
        else:
            index = cls.parsePropertyRef(stack, tokens, formula, index)
        return index

    @classmethod
    def parseArgs(cls, stack, tokens, formula, index):
        """
        Parse function element arguments.

        Args::
            args
                : conditionalOrExpression
                | conditionalOrExpression COMMA args
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        previous = None
        while index < len(tokens):
            index = cls.parseConditionalOrExpression(stack, tokens, formula, index)
            arg = stack[len(stack) - 1] if previous == None else stack.pop()
            if previous != None:
                previous.setRightHandSide(arg)
            previous = arg
            current = tokens[index]
            if current.getType() == FormulaParser.Token.Type.COMMA:
                index += 1
            else:
                return index
        return index

    @classmethod
    def parsePropertyRef(cls, stack, tokens, formula, index):
        """
        Parse a property reference.

        PropertyRef::
            propertyRef
                : DOLLAR? ATTRIBUTE IDENT RPAREN
                ;

        :param stack: the intermediate tokens of formula
        :param tokens: the formula tokens
        :param formula: the formula being parsed
        :param index: the current parse position in tokens
        :return: the index of the next token to be processed
        """
        if index >= len(tokens):
            return len(tokens)
        token = tokens[index]
        if token.getType() == FormulaParser.Token.Type.ATTRIBUTE or \
                token.getType() == FormulaParser.Token.Type.DOLLAR:
            current = token
            #  Handle attribute, which is $? $( IDENT )
            dollarCount = 0
            while current.getType() == cls.Token.Type.DOLLAR:
                dollarCount += 1
                if dollarCount > 1:
                    raise ArgumentException(
                        f'term: {current.getType()} @ '
                        f'{current.getPos()} not expected')
                index += 1
                current = cls.peek(tokens, index)
            # 0 is IN_PROCESS_ATTIRBUTE
            # 1 is CURRENT_ATTRIBUTE
            attrType = cls.Terminal.Type(dollarCount)
            if current.getType() != cls.Token.Type.ATTRIBUTE:
                raise ArgumentException(
                    f'term: Found {current.getType()} @ '
                    f'{current.getPos()}. Expected ATTRIBUTE')
            index += 1
            current = cls.peek(tokens, index)
            if current.getType() != cls.Token.Type.IDENT:
                raise ArgumentException(
                    f'term: Found {current.getType()} @ '
                    f'{current.getPos()}. Expected IDENT')
            value = formula[current.getPos():
                            current.getPos() + current.getLength()]
            index += 1
            current = cls.peek(tokens, index)
            if current.getType() != cls.Token.Type.RPAREN:
                raise ArgumentException(
                    f'term: Found {current.getType()} @ '
                    f'{current.getPos()}. Expected RPAREN')
            stack.append(cls.Terminal(attrType, value))
            index += 1
            #  consume RPAREN
        return index

    @classmethod
    def peek(cls, tokens, offset):
        """
        Return the token at offset.

        :param tokens: a `list` of :class:`.Token`
        :param offset: the integer offset into `tokens`
        :return: the :class:`.Token` at offset
        """
        index = offset if 0 <= offset and offset <= len(tokens) - 1 else len(tokens) - 1
        return tokens[index]


    @classmethod
    def prioritize(cls, lhs, rhs):
        """
        Prioritize `lhs` and `rhs` based on precedence.

        The left hand side needs to have higher precedence than right hand side
        so that post-fix traversal does higher precedence operations first.
        The swap on compare == 0 ensures the remaining operations are
        left-to-right.
        :param lhs: the left hand side `Node`
        :param rhs: the right hand side `Node`
        :return: the `Node` with higher precedence
        """
        if rhs.getOperation() != cls.Node.Operation.TERMINAL:
            c = cls.comparePrec(lhs, rhs)
            if c == 0:
                lhs.setRightHandSide(rhs.getLeftHandSide())
                rightHandSide = rhs.getRightHandSide()
                rhs.setLeftHandSide(lhs)
                rhs.setRightHandSide(rightHandSide)
                return rhs
            elif c > 0:
                leftHandSide = rhs.getLeftHandSide()
                rhs.setLeftHandSide(lhs)
                lhs.setRightHandSide(leftHandSide)
                return lhs
            else:
                lhs.setRightHandSide(rhs)
                return lhs
        else:
            lhs.setRightHandSide(rhs)
            return lhs

    @classmethod
    def dump(cls, node):
        """
        Return a string representation of `node`

        :param node: a `Node`
        :return: the string representation of `node`
        """
        if node is None:
            return None
        if isinstance(node, (cls.Terminal, )):
            s = node.getValue()
            if node.terminaltype == cls.Terminal.Type.IN_PROCESS_ATTRIBUTE:
                s = f'$({s})'
            elif node.terminaltype == cls.Terminal.Type.CURRENT_ATTRIBUTE:
                s = f'$$({s})'
            return s
        lhs = cls.dump(node.getLeftHandSide())
        rhs = cls.dump(node.getRightHandSide())
        operation = node.getOperation()
        return f'[{str(operation)}|{lhs}|{rhs}]'

# com.oracle.iot.shared

class ValueProvider(ABC):
    """
    Abstract a value in a formula.

    The value could be an "in-process" value or a current attribute value.
    """
    @abstractmethod
    def getCurrentValue(self, key):
        """
        Provides an attribute or other property for *$(propRef)* in a formula.
        :para key: the value of propRef in a formula
        :return: value of the given propRef or `None` if not found
        """

    @abstractmethod
    def getInProcessValue(self, key):
        """
        Provides an attribute or other property for *$$(propRef)* in a formula.
        :para key: the value of propRef in a formula
        :return: value of the given propRef or null if not found
        """
