# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC, abstractmethod
import json
import logging
import threading

#from iotcs.client.device.util import RequestDispatcher
from iotcs.common import Closeable
from iotcs.common import TimeManager
from iotcs.common import Util
from iotcs.exception import StateException
from iotcs.message import RequestMessage
from iotcs.message import ResponseMessage

# Until these are available
# Although this is not good given that they are at the top level

# com.oracle.iot.client.impl

class AccessToken:
    def __init__(self, expires, tokenType, token):
        self.expires = expires
        self.tokenType = tokenType
        self.token = token
        self.expirationTime = TimeManager.currentTimeMillis() + expires

    def hasExpired(self):
        return (TimeManager.currentTimeMillis() >= self.expirationTime)

    def getExpires(self):
        return self.expires

    def getTokenType(self):
        return self.tokenType

    def getToken(self):
        return self.token

    def toString(self):
        return "Accessor.Token{" + \
               "expires=" + self.expires + \
               ", tokenType='" + self.tokenType + "'" + \
               ", token='" + self.token + "'}"

    @staticmethod
    def fromJson(data):
        try:
            tokenJson = json.loads(data)
            return AccessToken(tokenJson["expires_in"], tokenJson["token_type"], tokenJson["access_token"])
        except Exception as exc:
            print(exc)
            return None

class SecureConnection(Closeable):
    """Provide secure request methods to an IoT server."""

    __slots__ = (
        'trustedAssetsManager',  # provides secure credentials
        'useOnlySharedSecret',  # use only a shared secret for a request
        'closed'  # set to `True` if connection is closed
    )

    def __init__(self, trustedAssetsManager, useOnlySharedSecret=False):
        """
        Init the `SecureConnection`.

        :param trustedAssetsManager: a
            :class:`iotcs.device.impl.UnifiedTrustedAssetsManager` instance
        :param useOnlySharedSecre: if `True` use only the shared secret for
            requests on this `SecureConnection`
            otherwise use signed credentials.
        """
        super(SecureConnection, self).__init__()
        self.trustedAssetsManager = trustedAssetsManager
        self.useOnlySharedSecret = useOnlySharedSecret

    def getEndpointId(self):
        """
        Return the device client identifier using this `SecureConnection`.

        :return: the device client identifier
        :raises: raises exceptions raised by
            :meth:`trustetAssetsMananger.getEndpointId` if
            :meth:`iotcs.device.impl.UnifiedTrustedAssetsManager.getClientId`
            returns none `None`.
        """
        try:
            return self.trustedAssetsManager.getEndpointId()
        except Exception as e:
            if self.trustedAssetsManager.getClientId() is None:
                return None
            raise e

    @abstractmethod
    def get(self, url, queryparams, headers=None, timeout=None):
        """
        Return the effective IoT server REST api URN.

        :param url:
        :param queryparams:
        :param headers:
        :param timeout:
        """
        pass

    @abstractmethod
    def post(self, url, queryparams, data, headers=None, timeout=None):
        """
        Send a `POST` message.

        :param url: the REST api resource URN
        :param queryParams:
        :param data: the message payload in json format
        :param headers:
        :param timeout: socket read timeout in milliseconds.
            timeout of zero is interpreted as an infinite timeout.
        """
        pass

    @abstractmethod
    def put(self, url, queryparams, data, headers=None, timeout=None):
        """
        Send a `PUT` message.

        :param url: the REST api resource URN
        :param queryparams:
        :param data: the message payload in json format
        :param headers:
        :param timeout:
        """
        pass

    @abstractmethod
    def delete(self, url, queryparams, data, headers=None, timeout=None):
        """
        Send a `DELETE` message.

        :param url: the REST api resource URN
        :param queryparams:
        :param data:
        :param headers:
        :param timeout:
        """
        pass

    @abstractmethod
    def patch(self, url, queryparams, data, headers=None, timeout=None):
        """
        Send a `PATCH` message.

        :param url: the REST api resource URN
        :param queryparams:
        :param data: the message payload in json format
        :param headers:
        :param timeout:
        """
        pass

    def close(self):
        """Close the connection."""
        self.closed = True
        self.disconnect()
        self.trustedAssetsManager.close()

    def isClosed(self):
        """Return `True` if this `SecureConnection` is closed, else `False`."""
        return self.closed

    @abstractmethod
    def disconnect(self):
        """Disconnect this `SecureConnection`."""
        pass

    def __eq__(self, other):
        """
        Return `True if `self` equals `other`.

        :param other: a `SecureConnection` instance
        """
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, (SecureConnection)):
            return False

        if not super(SecureConnection, self).__eq__(other):
            return False

        return (self.trustedAssetsManager.getServerScheme(),
                self.trustedAssetsManager.getServerHost(),
                self.trustedAssetsManager.getServerPort()) == \
            (other.trustedAssetsManager.getServerScheme(),
                other.trustedAssetsManager.getServerHost(),
                other.trustedAssetsManager.getServerPort())

    def __hash__(self):
        """Return a `hash` for this `SecureConnection`."""
        return hash((self.trustedAssetsManager.getServerScheme(),
                    self.trustedAssetsManager.getServerHost(),
                    self.trustedAssetsManager.getServerPort()))

    def getTrustedAssetsManager(self):
        """
        Return the `trustedAssetsManager for this `SecureConnection`.

        :return: `trustedAssetsManager`
        """
        return self.trustedAssetsManager

    def usesOnlySharedSecret(self):
        """
        Return `useOnlySharedSecret`.

        :return: `useOnlySharedSecret`
        """
        return self.useOnlySharedSecret


# head and tail are kept so that they are less than requestBuffer.length.
# This avoids the (unlikely) situation that they would overflow.
# But this also means that tail might be less than head, since the buffer
# is circular. This has to be taken into account when calculating the
# available bytes. Remember that tail is the pointer to the next
# available chunk and does not count toward used bytes (no need to
# adjust used byte count by 1). If tail == head, then used bytes is
# zero - perfect!
#


class RequestBuffer():
    """
    Circular buffer implementation for received messages.
    The buffer "blocks" contain a single json message. The first two bytes
    in a block are the length in bytes of the json message
    """

    # TODO: What about locking ?

    __slots__ = (
        'size',
        'buffer',
        'tail',
        'head',
        'messageCount',
        'bufferLock'
    )

    def __init__(self, size):
        self.size = size
        self.buffer = bytearray(self.size)
        self.head = 0
        self.tail = 0
        self.messageCount = 0
        self.bufferLock = threading.RLock()

    def getSize(self):
        return len(self.buffer)

    def getAvailableBytes(self):
        """
        Buffer space available for messages
        :return:
        """
        # subtract 2 for the length bytes.
        # TODO: subtracting 2 may fail if there is more than one message
        # coming from the server. Say we had 100 bytes available and we have two
        # messages coming of 49 bytes. We'd write 2 bytes for the 49, then the
        # 49 bytes. Now we only have 49 bytes of buffer left, which is not
        # enough to hold 2 bytes of length plus 49 bytes of message.
        with self.bufferLock:
            return len(self.buffer) - self.getUsedBytes() - 2

    def _getUsedBytes(self):
        with self.bufferLock:
            return self.tail - self.head if self.tail >= self.head else \
                (self.tail + len(self.buffer)) - self.head

    def getUsedBytes(self):
        return self._getUsedBytes()

    # def readDataLength(self):
    #     if self.getAvailableBytes() > 2:
    #         lenbuf = self.buffer[self.head:self.head+2]
    #         buflen = int.from_bytes(lenbuf,byteorder='big',
    #                                 signed=False)
    #         self.head += 2
    #         return buflen
    #     else:
    #         return None

    def read(self):
        """
        Dequeue a message by adjusting head and tail.
        :return: bytearray
        """
        with self.bufferLock:
            usedbytes = self._getUsedBytes()
            if usedbytes <= 2:
                return None

            lenbuf = self.buffer[self.head:self.head + 2]
            buflen = int.from_bytes(lenbuf, byteorder='big',
                                    signed=False)
            self.head += 2
            start = self.head

            if buflen <= 0:
                return None

            # the start of the message
            start = self.head

            # keep head < requestBuffer.length to avoid overflow of head.
            # this 'head' is also the end of the current message
            self.head = (self.head + buflen) % len(self.buffer)

            # return (pos < eos) ? buffer[(pos++ % buffer.length)] : -1;
            # does it wrap ?
            if start < self.head:
                # doesn' wrap return the message
                return self.buffer[start:start + buflen]

            # It wraps. The message it two parts from 'start' (the previous head + 2)
            # to the end of the buffer, and from the start of the buffer (0) to the
            # updated head
            readbuf = self.buffer[start:len(self.buffer)] + self.buffer[0:self.head]
            return readbuf

    def writeHeader(self, start, length):
        """
        Writes length bytes at the `start` of the message
        :param length:
        :return:
        """
        lengthbuf = length.to_bytes(2, byteorder='big')
        # TODO: Musn't this be 2 ?
        buflen = len(lengthbuf)
        self.buffer[start:start + buflen] = lengthbuf


    # TODO: Should this be external to make the RequestBuffer more generic ?
    # synchronized
    def write(self, data):
        """
        Assumes data is formatted as '[ {...} ]'
        :param data:
        :return:
        """

        if data is None or len(data) <= 2:
            return

        with self.bufferLock:
            #  if data.length == 2, then it is an empty json array and there are
            #  no values in the message.

            pos = 1
            data = str(data, 'utf-8')
            # TODO: What is "data" str or bytes
            while data[pos] == '{':
                start = pos
                msgstart = self.tail
                #  first two bytes is for length, start buffering json bytes at tail + 2
                self.tail += 2
                #  json objects start with '{', end with '}'
                #  count braces to pair '{' with matching '}'
                braceCount = 0
                while pos < len(data):
                    b = data[pos]
                    pos += 1
                    index = self.tail % len(self.buffer)
                    self.buffer[index] = ord(b)
                    self.tail += 1
                    if b == '{':
                        braceCount += 1
                        continue
                    if b == '}':
                        braceCount -= 1
                        if braceCount == 0:
                            break

                nbytes = pos - start

                # self.buffer[lengthIndex0++ % len(self.buffer)] = int(((0xff00 & nbytes) >> 8))
                # self.buffer[lengthIndex0 % len(self.buffer)] = int(((0x00ff & nbytes)))
                self.writeHeader(msgstart, nbytes)

                #  get past the end of the json object
                #  json objects end with '}' and json objects are separated by ','
                if pos < len(data) and data[pos] == '}':
                    pos += 1
                if pos < len(data) and data[pos] == ',':
                    pos += 1

                # If there was data, debug log it, else nothing to log.
                if pos != start:
                    self.getLogger().debug(f'buffered: {str(data[start:pos])}')

            #  adjust tail to keep it < buffer.length to avoid overflow.
            self.tail = self.tail % len(self.buffer)

            #  Logging done here since we need the adjusted tail to get the
            #  correct available bytes remaining in the buffer.

            #  subtract 2 for the length bytes.
            availableBytes = len(self.buffer) - self._getUsedBytes() - 2
            self.getLogger().debug(f'buffered {len(data)} bytes of request data from server. {availableBytes} available bytes of {len(self.buffer)} remaining in buffer.')


    def getLogger(self):
        return logging.getLogger('iotcs.shared.RequestBuffer')


class SendReceiveImpl(ABC):
    """
    `SendReceiveImpl` is an implementation of the `DirectlyConnectedDevice.send()`
    and `DirectlyConnectedDevice.receive()` methods.
    The send is synchronous. Receive should be considered as a synchronous call.
    The implementation has a buffer for receiving request messages. If there are
    no messages in the buffer, the receive implementation will send a message
    to the server to receive any pending requests the server may have.

    The size of the buffer can be configured by setting the property "request_buffer_size"
    in the "messaging" section of `iotcs.config.config.ini`.

    Messages received from server are handled using a circular buffer of
    bytes. The first two bytes at the head of the buffer always contains
    the number of bytes in the message that follows. The tail points to the
    next available chunk of buffer.  When bytes are written to the buffer,
    the bytes are written from (tail+2) to the end of the buffer, and then
    from buffer[0] if necessary. In this way, the buffer wraps from tail
    to head. Head keeps moving as messages are read. Tail keeps moving as
    requests are received. Bytes in the buffer never have to be moved.

    The number of bytes available are sent to the server in a query parameter
    to the messages REST API:
        POST /iot/api/v2/messages?acceptBytes=<number-of-bytes>
    """

    USE_DEFAULT_TIMEOUT_VALUE = -1
    REQUEST_BUFFER_SIZE_PROPERTY = "request_buffer_size"
    DISABLE_LONG_POLLING_PROPERTY = "disable_long_polling"
    LONG_POLLING_TIMEOUT_OFFSET_PROPERTY = "long_polling_timeout_offset"
    SEND_RECEIVE_TIMEOUT_PROPERTY = "send_receive_timeout"
    MIN_ACCEPT_BYTES_HEADER = "X-min-acceptBytes"

    __slots__ = (
        'requestBufferSize',
        'sendReceiveTimeLimit',
        'requestBuffer',
        'head',
        'tail',
        'sendCallTime',
        'sendCallTimeLock'
        'useLongPolling',
        'recvQueue',
        'receiveLock'
    )

    def __init__(self):
        config = Util.getConfig()
        self.requestBufferSize = config.getint(Util.CONFIG_DEVICE_MESSAGING_SECTION,
                                               self.REQUEST_BUFFER_SIZE_PROPERTY)
        self.sendReceiveTimeLimit = config.getint(Util.CONFIG_DEVICE_MESSAGING_SECTION,
                                                  self.SEND_RECEIVE_TIMEOUT_PROPERTY)
        self.requestBuffer = RequestBuffer(self.requestBufferSize)
        self.sendCallTime = -1
        self.sendCallTimeLock = threading.RLock()
        self.useLongPolling = not config.getboolean(Util.CONFIG_DEVICE_MESSAGING_SECTION,
                                                self.DISABLE_LONG_POLLING_PROPERTY)
        self.receiveLock = threading.RLock()

    #
    # This is where a call to DirectlyConnectedDevice#send(Message...)
    # ends up.
    #
    def send(self, messages):
        """

        :param messages: A list of `Message`
        :return:
        :raises: :class:`.StateException` if message cannot be buffered
        """

        if messages is None or len(messages) <= 0:
            return

        # Json array
        payload = []
        for message in messages:
            # Special handling for LL actionCondition
            #     If this is a request message that loops back to the sender,
            #         don't deliver it, but queue it up for the next call to receive.
            #     If this is a response message that loops back to the sender,
            #         then don't send it to the server.
            if isinstance(message, RequestMessage):
                if message.getDestination() is not None and \
                        message.getDestination() == message.getSource():
                    jsonObject = message.toJson()
                    jsonString = f'[{json.dumps(jsonObject).decode()}]'
                    try:
                        self.requestBuffer.write(jsonString)
                        continue
                    except Exception as e:
                        # Use IOError like JCL uses IOException
                        raise IOError(f'SendReceiveImpl.send(): '
                            f'could not loopback request message {e}')

            elif isinstance(message, ResponseMessage):
                # "loopback response" Where do these come from ?
                if message.getDestination() is not None and \
                        message.getDestination() == message.getSource():
                    continue

            jsonObject = message.toJson()
            payload.append(jsonObject)

        if len(payload) == 0:
            return

        self.post(payload)

        if not self.useLongPolling:
            with self.sendCallTimeLock:
                self.sendCallTime = TimeManager.currentTimeMillis()

        return

    @abstractmethod
    def post(self, payload, timeout=USE_DEFAULT_TIMEOUT_VALUE):
        """
        Messages are received when posting, an application when using HTTP long
        polling can set the read timeout (in milliseconds) when receiving,
        the so it is only need for the `HttpSendReceiveImpl`,
        which passes it to `HttpSecureConnection`, which passes it to `HttpClient`,
        which passes it to `HttpsUrlConnection.setReadTimeout`, in which
        timeout of zero is interpreted as an infinite timeout.

        However, if a negative value is given, it will be converted to be the
        default response timeout.

        :param payload:
        :param timeout:
        :return:
        """
        pass

    # synchronized
    def receive(self, timeout):
        """
        Originates from  `DirectlyConnectedDevice.receive()`.

        `timeout` (in milliseconds) is used when using HTTP long polling,
        so it is only needed for the `HttpSendReceiveImpl` requests,
        which passes it to `HttpSecureConnection`, which passes it to `HttpClient`,
        which passes it to <WHAT IS THE URLLIB3 API FOR SETTING TIMEOUT>
        A `timeout` of zero is interpreted as an infinite timeout.

        However, if a negative value is given, it will be converted to be the
        default response timeout.

        Get request messages from the IOT Cloud Service
        :param timeout:
        :return: `RequestMessage`
        """

        with self.receiveLock:
            # If there are no pending request messages make a
            # server request to see if there are any on the server
            jsonMessage = self.requestBuffer.read()
            if jsonMessage is None:
                if not self.useLongPolling:
                    sendcalltime = 0
                    with self.sendCallTimeLock:
                        sendcalltime = self.sendCallTime

                    receiveCallTime = TimeManager.currentTimeMillis()
                    if self.sendReceiveTimeLimit == 0 or \
                            (receiveCallTime - sendcalltime) < self.sendReceiveTimeLimit:
                        # time delta between last send and this receive is too
                        # small do not make a call to the network to get request
                        # messages
                        return None
                    else:
                        self.post(None, self.USE_DEFAULT_TIMEOUT_VALUE)
                        self.sendCallTime = TimeManager.currentTimeMillis()
                else:
                    # May add to the request buffer so read again on return
                    self.post(None, timeout)

            # If the first read yielded a message, then jsonMessage is not None
            # otherwise read from the
            if jsonMessage is None:
                jsonMessage = self.requestBuffer.read()

            if jsonMessage is not None:
                self.getLogger().debug(f'receive(): dequeued: {jsonMessage}')
                requestMessage = RequestMessage().fromJson(json.loads(jsonMessage))
                return requestMessage
            else:
                return None


    def getLogger(self):
        return logging.getLogger('iotcs.device.impl.SendReceiveImpl')
